""" "
Re-entrancy detection

Based on heuristics, it may lead to FP and FN
Iterate over all the nodes of the graph until reaching a fixpoint
"""

from collections import namedtuple, defaultdict

from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.reentrancy.reentrancy import Reentrancy, to_hashable
from slither.utils.output import Output

FindingKey = namedtuple("FindingKey", ["function", "calls", "send_eth"])
FindingValue = namedtuple("FindingValue", ["variable", "node", "nodes"])


class ReentrancyBenign(Reentrancy):
    ARGUMENT = "reentrancy-benign"
    HELP = "Benign reentrancy vulnerabilities"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#reentrancy-vulnerabilities-3"
    )

    WIKI_TITLE = "Reentrancy vulnerabilities"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detection of the [reentrancy bug](https://github.com/trailofbits/not-so-smart-contracts/tree/master/reentrancy).
Only report reentrancy that acts as a double call (see `reentrancy-eth`, `reentrancy-no-eth`)."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    function callme(){
        if( ! (msg.sender.call()() ) ){
            throw;
        }
        counter += 1
    }
```

`callme` contains a reentrancy. The reentrancy is benign because it's exploitation would have the same effect as two consecutive calls."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Apply the [`check-effects-interactions` pattern](http://solidity.readthedocs.io/en/v0.4.21/security-considerations.html#re-entrancy)."

    STANDARD_JSON = False

    def find_reentrancies(self) -> defaultdict[FindingKey, set[FindingValue]]:
        result = defaultdict(set)
        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                for node in f.nodes:
                    # dead code
                    if self.KEY not in node.context:
                        continue
                    if node.context[self.KEY].calls:
                        if not any(n != node for n in node.context[self.KEY].calls):
                            continue
                        read_then_written = []
                        for c in node.context[self.KEY].calls:
                            read_then_written += [
                                v
                                for v in node.context[self.KEY].written
                                if v in node.context[self.KEY].reads_prior_calls[c]
                            ]
                        not_read_then_written = {
                            FindingValue(
                                v,
                                node,
                                tuple(sorted(nodes, key=lambda x: x.node_id)),
                            )
                            for (v, nodes) in node.context[self.KEY].written.items()
                            if v not in read_then_written
                        }
                        if not_read_then_written:
                            # calls are ordered
                            finding_key = FindingKey(
                                function=node.function,
                                calls=to_hashable(node.context[self.KEY].calls),
                                send_eth=to_hashable(node.context[self.KEY].send_eth),
                            )
                            result[finding_key] |= not_read_then_written
        return result

    def _detect(self) -> list[Output]:
        """"""

        super()._detect()
        reentrancies = self.find_reentrancies()

        results = []

        result_sorted = sorted(reentrancies.items(), key=lambda x: x[0].function.name)
        varsWritten: list[FindingValue]
        for (func, calls, send_eth), varsWritten in result_sorted:
            calls = sorted(set(calls), key=lambda x: x[0].node_id)
            send_eth = sorted(set(send_eth), key=lambda x: x[0].node_id)
            varsWritten = sorted(varsWritten, key=lambda x: (x.variable.name, x.node.node_id))

            info = ["Reentrancy in ", func, ":\n"]

            info += ["\tExternal calls:\n"]
            for call_info, calls_list in calls:
                info += ["\t- ", call_info, "\n"]
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        info += ["\t\t- ", call_list_info, "\n"]
            if calls != send_eth and send_eth:
                info += ["\tExternal calls sending eth:\n"]
                for call_info, calls_list in send_eth:
                    info += ["\t- ", call_info, "\n"]
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            info += ["\t\t- ", call_list_info, "\n"]
            info += ["\tState variables written after the call(s):\n"]
            for finding_value in varsWritten:
                info += ["\t- ", finding_value.node, "\n"]
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        info += ["\t\t- ", other_node, "\n"]

            # Create our JSON result
            res = self.generate_result(info)

            # Add the function with the re-entrancy first
            res.add(func)

            # Add all underlying calls in the function which are potentially problematic.
            for call_info, calls_list in calls:
                res.add(call_info, {"underlying_type": "external_calls"})
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        res.add(
                            call_list_info,
                            {"underlying_type": "external_calls_sending_eth"},
                        )

            #

            # If the calls are not the same ones that send eth, add the eth sending nodes.
            if calls != send_eth:
                for call_info, calls_list in calls:
                    res.add(call_info, {"underlying_type": "external_calls_sending_eth"})
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            res.add(
                                call_list_info,
                                {"underlying_type": "external_calls_sending_eth"},
                            )

            # Add all variables written via nodes which write them.
            for finding_value in varsWritten:
                res.add(
                    finding_value.node,
                    {
                        "underlying_type": "variables_written",
                        "variable_name": finding_value.variable.name,
                    },
                )
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        res.add(
                            other_node,
                            {
                                "underlying_type": "variables_written",
                                "variable_name": finding_value.variable.name,
                            },
                        )

            # Append our result
            results.append(res)

        return results



from shutil import get_terminal_size

from prettytable import PrettyTable
from prettytable.colortable import ColorTable, Themes

from slither.utils.colors import Colors


class MyPrettyTable:
    def __init__(
        self,
        field_names: list[str],
        pretty_align: bool = True,
        max_width: int | None = "max",  # Default value is "max"
    ):
        self._field_names = field_names
        self._rows: list = []
        self._options: dict = {}
        if pretty_align:
            self._options["set_alignment"] = []
            for field_name in field_names:
                self._options["set_alignment"] += [(field_name, "l")]
        else:
            self._options["set_alignment"] = []

        self.max_width = None
        if max_width == "max":
            # We use (0,0) as a fallback to detect if we are not attached to a terminal
            # In this case, we fall back to the default behavior (i.e. printing as much as possible)
            terminal_column = get_terminal_size((0, 0)).columns
            if terminal_column != 0:
                # We reduce slightly the max-width to take into account inconsistencies in terminals
                self.max_width = terminal_column - 3
        else:
            self.max_width = max_width

    def add_row(self, row: list[str | list[str]]) -> None:
        self._rows.append(row)

    def to_pretty_table(self) -> PrettyTable:
        if Colors.COLORIZATION_ENABLED:
            table = ColorTable(self._field_names, theme=Themes.OCEAN)
        else:
            table = PrettyTable(self._field_names)

        if self.max_width is not None:
            table.max_table_width = self.max_width

        for row in self._rows:
            table.add_row(row)
        if len(self._options["set_alignment"]):
            for column_header, value in self._options["set_alignment"]:
                table.align[column_header] = value
        return table

    def to_json(self) -> dict:
        return {"fields_names": self._field_names, "rows": self._rows}

    def __str__(self) -> str:
        return str(self.to_pretty_table())


# UTILITY FUNCTIONS


def make_pretty_table(
    headers: list, body: dict, totals: bool = False, total_header="TOTAL"
) -> MyPrettyTable:
    """
    Converts a dict to a MyPrettyTable.  Dict keys are the row headers.
    Args:
        headers: str[] of column names
        body: dict of row headers with a dict of the values
        totals: bool optional add Totals row
        total_header: str optional if totals is set to True this will override the default "TOTAL" header
    Returns:
        MyPrettyTable
    """
    table = MyPrettyTable(headers)
    for row in body:
        table_row = [row] + [body[row][key] for key in headers[1:]]
        table.add_row(table_row)
    if totals:
        table.add_row([total_header] + [sum(body[row][key] for row in body) for key in headers[1:]])
    return table



"""
Module detecting EIP-2612 domain separator collision
"""

from slither.core.declarations import Function
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.function import get_function_id
from slither.utils.output import Output


class DomainSeparatorCollision(AbstractDetector):
    """
    Domain separator collision
    """

    ARGUMENT = "domain-separator-collision"
    HELP = "Detects ERC20 tokens that have a function whose signature collides with EIP-2612's DOMAIN_SEPARATOR()"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#domain-separator-collision"
    )

    WIKI_TITLE = "Domain separator collision"
    WIKI_DESCRIPTION = "An ERC20 token has a function whose signature collides with EIP-2612's DOMAIN_SEPARATOR(), causing unanticipated behavior for contracts using `permit` functionality."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    function some_collisions() external() {}
}
```
`some_collision` clashes with EIP-2612's DOMAIN_SEPARATOR() and will interfere with contract's using `permit`."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove or rename the function that collides with DOMAIN_SEPARATOR()."

    def _detect(self) -> list[Output]:
        domain_sig = get_function_id("DOMAIN_SEPARATOR()")
        for contract in self.compilation_unit.contracts_derived:
            if contract.is_erc20():
                funcs_and_vars: list[Function | StateVariable] = (
                    contract.functions_entry_points + contract.state_variables_entry_points
                )  # type: ignore
                for func_or_var in funcs_and_vars:
                    # External/ public function names should not collide with DOMAIN_SEPARATOR()
                    hash_collision = (
                        func_or_var.solidity_signature != "DOMAIN_SEPARATOR()"
                        and get_function_id(func_or_var.solidity_signature) == domain_sig
                    )
                    # DOMAIN_SEPARATOR() should return bytes32
                    incorrect_return_type = func_or_var.solidity_signature == "DOMAIN_SEPARATOR()"
                    if incorrect_return_type:
                        if isinstance(func_or_var, Function):
                            incorrect_return_type = (
                                not func_or_var.return_type
                                or func_or_var.return_type[0] != ElementaryType("bytes32")
                            )
                        else:
                            assert isinstance(func_or_var, StateVariable)
                            incorrect_return_type = func_or_var.type != ElementaryType("bytes32")
                    if hash_collision or incorrect_return_type:
                        info: DETECTOR_INFO = [
                            "The function signature of ",
                            func_or_var,
                            " collides with DOMAIN_SEPARATOR and should be renamed or removed.\n",
                        ]
                        res = self.generate_result(info)
                        return [res]
        return []



from pathlib import Path
from crytic_compile import CryticCompile
from crytic_compile.platform.solc_standard_json import SolcStandardJson

from slither import Slither
from slither.slithir.operations import InternalCall, LibraryCall

from tests.utils import _run_all_detectors

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
USING_FOR_TEST_DATA_DIR = Path(TEST_DATA_DIR, "using_for")


def test_using_for_global_collision(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    standard_json = SolcStandardJson()
    for source_file in Path(USING_FOR_TEST_DATA_DIR, "using_for_global_collision").rglob("*.sol"):
        standard_json.add_source_file(Path(source_file).as_posix())
    compilation = CryticCompile(standard_json, solc=solc_path)
    sl = Slither(compilation, disallow_partial=True)
    _run_all_detectors(sl)


def test_using_for_top_level_same_name(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(USING_FOR_TEST_DATA_DIR, "using-for-3-0.8.0.sol").as_posix(), solc=solc_path
    )
    contract_c = slither.get_contract_from_name("C")[0]
    libCall = contract_c.get_function_from_full_name("libCall(uint256)")
    found = False
    for ir in libCall.all_slithir_operations():
        if isinstance(ir, LibraryCall) and ir.destination == "Lib" and ir.function_name == "a":
            found = True
    assert found


def test_using_for_top_level_implicit_conversion(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(USING_FOR_TEST_DATA_DIR, "using-for-4-0.8.0.sol").as_posix(), solc=solc_path
    )
    contract_c = slither.get_contract_from_name("C")[0]
    libCall = contract_c.get_function_from_full_name("libCall(uint16)")
    found = False
    for ir in libCall.all_slithir_operations():
        if isinstance(ir, LibraryCall) and ir.destination == "Lib" and ir.function_name == "f":
            found = True
    assert found


def test_using_for_alias_top_level(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(USING_FOR_TEST_DATA_DIR, "using-for-alias-top-level-0.8.0.sol").as_posix(),
        solc=solc_path,
    )
    contract_c = slither.get_contract_from_name("C")[0]
    libCall = contract_c.get_function_from_full_name("libCall(uint256)")
    found = False
    for ir in libCall.all_slithir_operations():
        if isinstance(ir, LibraryCall) and ir.destination == "Lib" and ir.function_name == "b":
            found = True
    assert found

    found = False
    topLevelCall = contract_c.get_function_from_full_name("topLevel(uint256)")
    for ir in topLevelCall.all_slithir_operations():
        if isinstance(ir, InternalCall) and ir.function_name == "a":
            found = True
    assert found


def test_using_for_alias_contract(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(USING_FOR_TEST_DATA_DIR, "using-for-alias-contract-0.8.0.sol").as_posix(),
        solc=solc_path,
    )
    contract_c = slither.get_contract_from_name("C")[0]
    libCall = contract_c.get_function_from_full_name("libCall(uint256)")
    found = False
    for ir in libCall.all_slithir_operations():
        if isinstance(ir, LibraryCall) and ir.destination == "Lib" and ir.function_name == "b":
            found = True

    assert found

    found = False
    topLevelCall = contract_c.get_function_from_full_name("topLevel(uint256)")
    for ir in topLevelCall.all_slithir_operations():
        if isinstance(ir, InternalCall) and ir.function_name == "a":
            found = True
    assert found


def test_using_for_in_library(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(USING_FOR_TEST_DATA_DIR, "using-for-in-library-0.8.0.sol").as_posix(), solc=solc_path
    )
    contract_c = slither.get_contract_from_name("A")[0]
    libCall = contract_c.get_function_from_full_name("a(uint256)")
    found = False
    for ir in libCall.all_slithir_operations():
        if isinstance(ir, LibraryCall) and ir.destination == "B" and ir.function_name == "b":
            found = True
    assert found


def test_using_for_constant_folding(slither_from_solidity_source) -> None:
    # https://github.com/crytic/slither/issues/2307
    source = """
            library SafeMath {
            uint256 private constant twelve = 12;
            struct A {uint256 a;}
            function add(A[twelve] storage z) internal { }
        }

        contract MathContract {
            uint256 private constant twelve = 12;
            using SafeMath for SafeMath.A[twelve];
            SafeMath.A[twelve] public z;
            function safeAdd() public {
                z.add();
            }
        }
    """
    with slither_from_solidity_source(source) as slither:
        contract = slither.get_contract_from_name("MathContract")[0]
        add = contract.get_function_from_full_name("safeAdd()")
        found = False
        for ir in add.all_slithir_operations():
            if isinstance(ir, LibraryCall) and ir.function_name == "add":
                found = True
        assert found



from typing import Optional, TYPE_CHECKING

from slither.core.declarations.top_level import TopLevel
from slither.core.variables.variable import Variable

if TYPE_CHECKING:
    from slither.core.cfg.node import Node
    from slither.core.scope.scope import FileScope


class TopLevelVariable(TopLevel, Variable):
    def __init__(self, scope: "FileScope") -> None:
        super().__init__()
        self._node_initialization: Node | None = None
        self.file_scope = scope

    # endregion
    ###################################################################################
    ###################################################################################
    # region IRs (initialization)
    ###################################################################################
    ###################################################################################

    @property
    def node_initialization(self) -> Optional["Node"]:
        """
        Node for the state variable initalization
        :return:
        """
        return self._node_initialization

    @node_initialization.setter
    def node_initialization(self, node_initialization):
        self._node_initialization = node_initialization

    # endregion
    ###################################################################################
    ###################################################################################



import sys

from slither.core.declarations.solidity_variables import SolidityVariableComposed
from slither.core.variables.state_variable import StateVariable
from slither.slither import Slither
from slither.slithir.operations.high_level_call import HighLevelCall
from slither.slithir.operations.index import Index
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.temporary import TemporaryVariable


def visit_node(node, visited):
    if node is None:
        return

    if node in visited:
        return

    visited += [node]
    taints = node.function.compilation_unit.context[KEY]

    refs = {}
    for ir in node.irs:
        if isinstance(ir, Index):
            refs[ir.lvalue] = ir.variable_left

        if isinstance(ir, Index):
            read = [ir.variable_left]
        else:
            read = ir.read
        print(ir)
        print(f"Refs {refs}")
        print(f"Read {[str(x) for x in ir.read]}")
        print(f"Before {[str(x) for x in taints]}")
        if any(var_read in taints for var_read in read):
            taints += [ir.lvalue]
            lvalue = ir.lvalue
            while isinstance(lvalue, ReferenceVariable):
                taints += [refs[lvalue]]
                lvalue = refs[lvalue]

        print(f"After {[str(x) for x in taints]}")
        print()

    taints = [v for v in taints if not isinstance(v, (TemporaryVariable, ReferenceVariable))]

    node.function.compilation_unit.context[KEY] = list(set(taints))

    for son in node.sons:
        visit_node(son, visited)


def check_call(func, taints):
    for node in func.nodes:
        for ir in node.irs:
            if isinstance(ir, HighLevelCall):
                if ir.destination in taints:
                    print(f"Call to tainted address found in {function.name}")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("python taint_mapping.py taint.sol")
        sys.exit(-1)

    # Init slither
    slither = Slither(sys.argv[1])

    initial_taint = [SolidityVariableComposed("msg.sender")]
    initial_taint += [SolidityVariableComposed("msg.value")]

    KEY = "TAINT"

    prev_taints = []
    slither.context[KEY] = initial_taint
    while set(prev_taints) != set(slither.context[KEY]):
        prev_taints = slither.context[KEY]
        for contract in slither.contracts:
            for function in contract.functions:
                print(f"Function {function.name}")
                slither.context[KEY] = list(set(slither.context[KEY] + function.parameters))
                visit_node(function.entry_point, [])
                print(f"All variables tainted : {[str(v) for v in slither.context[KEY]]}")

            for function in contract.functions:
                check_call(function, slither.context[KEY])

    print(
        f"All state variables tainted : {[str(v) for v in prev_taints if isinstance(v, StateVariable)]}"
    )



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class Backdoor(AbstractDetector):
    """
    Detect function named backdoor
    """

    ARGUMENT = "backdoor"  # slither will launch the detector with slither.py --mydetector
    HELP = "Function named backdoor (detector example)"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/trailofbits/slither/wiki/Adding-a-new-detector"
    WIKI_TITLE = "Backdoor example"
    WIKI_DESCRIPTION = "Plugin example"
    WIKI_EXPLOIT_SCENARIO = ".."
    WIKI_RECOMMENDATION = ".."

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            # Check if a function has 'backdoor' in its name
            for f in contract.functions:
                if "backdoor" in f.name:
                    # Info to be printed
                    info: DETECTOR_INFO = ["Backdoor function found in ", f, "\n"]

                    # Add the result in result
                    res = self.generate_result(info)

                    results.append(res)

        return results



from typing import TYPE_CHECKING

from slither.core.declarations import Enum
from slither.core.declarations.top_level import TopLevel

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope


class EnumTopLevel(Enum, TopLevel):
    def __init__(
        self, name: str, canonical_name: str, values: list[str], scope: "FileScope"
    ) -> None:
        super().__init__(name, canonical_name, values)
        self.file_scope: FileScope = scope



"""
Check if ethers are locked in the contract
"""

from slither.core.declarations import Contract, SolidityFunction
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import (
    HighLevelCall,
    LowLevelCall,
    Send,
    Transfer,
    NewContract,
    LibraryCall,
    InternalCall,
    SolidityCall,
)
from slither.slithir.variables import Constant
from slither.utils.output import Output


class LockedEther(AbstractDetector):
    ARGUMENT = "locked-ether"
    HELP = "Contracts that lock ether"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#contracts-that-lock-ether"

    WIKI_TITLE = "Contracts that lock Ether"
    WIKI_DESCRIPTION = "Contract with a `payable` function, but without a withdrawal capacity."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
pragma solidity 0.4.24;
contract Locked{
    function receive() payable public{
    }
}
```
Every Ether sent to `Locked` will be lost."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove the payable attribute or add a withdraw function."

    @staticmethod
    def do_no_send_ether(contract: Contract) -> bool:
        functions = contract.all_functions_called
        to_explore = functions
        explored = []
        while to_explore:
            functions = to_explore
            explored += to_explore
            to_explore = []
            for function in functions:
                calls = [ir.function.name for ir in function.internal_calls]
                if "suicide(address)" in calls or "selfdestruct(address)" in calls:
                    return False
                for node in function.nodes:
                    for ir in node.irs:
                        if isinstance(
                            ir,
                            (Send, Transfer, HighLevelCall, LowLevelCall, NewContract),
                        ):
                            if ir.call_value and ir.call_value != 0:
                                return False
                        if isinstance(ir, (LowLevelCall)) and ir.function_name in [
                            "delegatecall",
                            "callcode",
                        ]:
                            return False
                        if isinstance(ir, SolidityCall):
                            call_can_send_ether = ir.function in [
                                SolidityFunction(
                                    "delegatecall(uint256,uint256,uint256,uint256,uint256,uint256)"
                                ),
                                SolidityFunction(
                                    "callcode(uint256,uint256,uint256,uint256,uint256,uint256,uint256)"
                                ),
                                SolidityFunction(
                                    "call(uint256,uint256,uint256,uint256,uint256,uint256,uint256)"
                                ),
                            ]
                            nonzero_call_value = call_can_send_ether and (
                                not isinstance(ir.arguments[2], Constant)
                                or ir.arguments[2].value != 0
                            )
                            if nonzero_call_value:
                                return False
                        # If a new internal call or librarycall
                        # Add it to the list to explore
                        # InternalCall if to follow internal call in libraries
                        if isinstance(ir, (InternalCall, LibraryCall)):
                            if ir.function not in explored:
                                to_explore.append(ir.function)

        return True

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            if contract.is_signature_only():
                continue
            funcs_payable = [function for function in contract.functions if function.payable]
            if funcs_payable:
                if self.do_no_send_ether(contract):
                    info: DETECTOR_INFO = ["Contract locking ether found:\n"]
                    info += ["\tContract ", contract, " has payable functions:\n"]
                    for function in funcs_payable:
                        info += ["\t - ", function, "\n"]
                    info += "\tBut does not have a function to withdraw the ether\n"

                    json = self.generate_result(info)

                    results.append(json)

        return results



"""
Module printing summary of the contract
"""

from slither.core.declarations import Function
from slither.printers.abstract_printer import AbstractPrinter


def _print_function(function: Function) -> str:
    txt = ""
    for node in function.nodes:
        if node.expression:
            txt += f"\t\tExpression: {node.expression}\n"
            txt += "\t\tIRs:\n"
            for ir in node.irs:
                txt += f"\t\t\t{ir}\n"
        elif node.irs:
            txt += "\t\tIRs:\n"
            for ir in node.irs:
                txt += f"\t\t\t{ir}\n"
    return txt


class PrinterSlithIR(AbstractPrinter):
    ARGUMENT = "slithir"
    HELP = "Print the slithIR representation of the functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#slithir"

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        for compilation_unit in self.slither.compilation_units:
            for contract in compilation_unit.contracts:
                txt += f"Contract {contract.name}\n"
                for function in contract.functions:
                    txt += f"\tFunction {function.canonical_name} {'' if function.is_shadowed else '(*)'}\n"
                    txt += _print_function(function)
                for modifier in contract.modifiers:
                    txt += f"\tModifier {modifier.canonical_name}\n"
                    txt += _print_function(modifier)
            if compilation_unit.functions_top_level:
                txt += "Top level functions\n"
            for function in compilation_unit.functions_top_level:
                txt += f"\tFunction {function.canonical_name}\n"
                txt += _print_function(function)
        self.info(txt)
        res = self.generate_output(txt)
        return res



from slither.tools.properties.properties.properties import (
    Property,
    PropertyType,
    PropertyReturn,
    PropertyCaller,
)

ERC20_CONFIG = [
    Property(
        name="init_total_supply()",
        description="The total supply is correctly initialized.",
        content="""
\t\treturn this.totalSupply() >= 0 && this.totalSupply() == initialTotalSupply;""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="init_owner_balance()",
        description="Owner's balance is correctly initialized.",
        content="""
\t\treturn initialBalance_owner == this.balanceOf(crytic_owner);""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="init_user_balance()",
        description="User's balance is correctly initialized.",
        content="""
\t\treturn initialBalance_user == this.balanceOf(crytic_user);""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="init_attacker_balance()",
        description="Attacker's balance is correctly initialized.",
        content="""
\t\treturn initialBalance_attacker == this.balanceOf(crytic_attacker);""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ANY,
    ),
    Property(
        name="init_caller_balance()",
        description="All the users have a positive balance.",
        content="""
\t\treturn this.balanceOf(msg.sender) >0 ;""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ALL,
    ),
    # Note: there is a potential overflow on the addition, but we dont consider it
    Property(
        name="init_total_supply_is_balances()",
        description="The total supply is the user and owner balance.",
        content="""
\t\treturn this.balanceOf(crytic_owner) + this.balanceOf(crytic_user) + this.balanceOf(crytic_attacker) == this.totalSupply();""",
        type=PropertyType.CODE_QUALITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=False,
        caller=PropertyCaller.ANY,
    ),
]



"""
Module detecting any path leading to usage of a local variable before it is declared.
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.variables.local_variable import LocalVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class PredeclarationUsageLocal(AbstractDetector):
    """
    Pre-declaration usage of local variable
    """

    ARGUMENT = "variable-scope"
    HELP = "Local variables used prior their declaration"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#pre-declaration-usage-of-local-variables"

    WIKI_TITLE = "Pre-declaration usage of local variables"
    WIKI_DESCRIPTION = "Detects the possible usage of a variable before the declaration is stepped over (either because it is later declared, or declared in another scope)."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    function f(uint z) public returns (uint) {
        uint y = x + 9 + z; // 'x' is used pre-declaration
        uint x = 7;

        if (z % 2 == 0) {
            uint max = 5;
            // ...
        }

        // 'max' was intended to be 5, but it was mistakenly declared in a scope and not assigned (so it is zero).
        for (uint i = 0; i < max; i++) {
            x += 1;
        }

        return x;
    }
}
```
In the case above, the variable `x` is used before its declaration, which may result in unintended consequences.
Additionally, the for-loop uses the variable `max`, which is declared in a previous scope that may not always be reached. This could lead to unintended consequences if the user mistakenly uses a variable prior to any intended declaration assignment. It also may indicate that the user intended to reference a different variable."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Move all variable declarations prior to any usage of the variable, and ensure that reaching a variable declaration does not depend on some conditional if it is used unconditionally."

    VULNERABLE_SOLC_VERSIONS = ALL_SOLC_VERSIONS_04

    def detect_predeclared_local_usage(
        self,
        node: Node,
        results: list[tuple[Node, LocalVariable]],
        already_declared: set[LocalVariable],
        visited: set[Node],
    ) -> None:
        """
        Detects if a given node uses a variable prior to declaration in any code path.
        :param node: The node to initiate the scan from (searches recursively through all sons)
        :param already_declared: A set of variables already known to be declared in this path currently.
        :param already_visited: A set of nodes already visited in this path currently.
        :param results: A list of tuple(node, local_variable) denoting nodes which used a variable before declaration.
        :return: None
        """

        if node in visited:
            return

        visited = visited | {node}

        if node.variable_declaration:
            already_declared = already_declared | {node.variable_declaration}

        if node not in self.fix_point_information:
            self.fix_point_information[node] = []

        # If we already explored this node with the same information
        if already_declared:
            for fix_point in self.fix_point_information[node]:
                if fix_point == already_declared:
                    return

        if already_declared:
            self.fix_point_information[node] += [already_declared]

        for variable in set(node.local_variables_read + node.local_variables_written):
            if variable not in already_declared:
                result = (node, variable)
                if result not in results:
                    results.append(result)

        for son in node.sons:
            self.detect_predeclared_local_usage(son, results, already_declared, visited)

    def detect_predeclared_in_contract(
        self, contract: Contract
    ) -> list[tuple[Function, list[tuple[Node, LocalVariable]]]]:
        """
        Detects and returns all nodes in a contract which use a variable before it is declared.
        :param contract: Contract to detect pre-declaration usage of locals within.
        :return: A list of tuples: (function, list(tuple(node, local_variable)))
        """

        # Create our result set.
        results: list[tuple[Function, list[tuple[Node, LocalVariable]]]] = []

        # Loop for each function and modifier's nodes and analyze for predeclared local variable usage.
        for function in contract.functions_and_modifiers_declared:
            predeclared_usage: list[tuple[Node, LocalVariable]] = []
            if function.nodes:
                self.detect_predeclared_local_usage(
                    function.nodes[0],
                    predeclared_usage,
                    set(function.parameters + function.returns),
                    set(),
                )
            if predeclared_usage:
                results.append((function, predeclared_usage))

        # Return the resulting set of nodes which set array length.
        return results

    def _detect(self) -> list[Output]:
        """
        Detect usage of a local variable before it is declared.
        """
        results = []

        # Fix_point_information contains a list of set
        # Each set contains the already declared variables saw in one path
        # If a path has the same set as a path already explored
        # We don't need to continue

        self.fix_point_information = {}

        for contract in self.contracts:
            predeclared_usages = self.detect_predeclared_in_contract(contract)
            if predeclared_usages:
                for predeclared_usage_function, predeclared_usage_nodes in predeclared_usages:
                    for (
                        predeclared_usage_node,
                        predeclared_usage_local_variable,
                    ) in predeclared_usage_nodes:
                        info: DETECTOR_INFO = [
                            "Variable '",
                            predeclared_usage_local_variable,
                            "' in ",
                            predeclared_usage_function,
                            " potentially used before declaration: ",
                            predeclared_usage_node,
                            "\n",
                        ]

                        res = self.generate_result(info)
                        results.append(res)

        return results



"""
Module detecting reserved keyword shadowing
"""

from slither.core.declarations import Function, Event
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.modifier import Modifier
from slither.core.variables import Variable
from slither.core.variables.local_variable import LocalVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class BuiltinSymbolShadowing(AbstractDetector):
    """
    Built-in symbol shadowing
    """

    ARGUMENT = "shadowing-builtin"
    HELP = "Built-in symbol shadowing"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#builtin-symbol-shadowing"

    WIKI_TITLE = "Built-in Symbol Shadowing"
    WIKI_DESCRIPTION = "Detection of shadowing built-in symbols using local variables, state variables, functions, modifiers, or events."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
pragma solidity ^0.4.24;

contract Bug {
    uint now; // Overshadows current time stamp.

    function assert(bool condition) public {
        // Overshadows built-in symbol for providing assertions.
    }

    function get_next_expiration(uint earlier_time) private returns (uint) {
        return now + 259200; // References overshadowed timestamp.
    }
}
```
`now` is defined as a state variable, and shadows with the built-in symbol `now`. The function `assert` overshadows the built-in `assert` function. Any use of either of these built-in symbols may lead to unexpected results."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Rename the local variables, state variables, functions, modifiers, and events that shadow a built-in symbol."

    SHADOWING_FUNCTION = "function"
    SHADOWING_MODIFIER = "modifier"
    SHADOWING_LOCAL_VARIABLE = "local variable"
    SHADOWING_STATE_VARIABLE = "state variable"
    SHADOWING_EVENT = "event"

    # Reserved keywords reference: https://solidity.readthedocs.io/en/v0.5.2/units-and-global-variables.html
    BUILTIN_SYMBOLS = [
        "assert",
        "require",
        "revert",
        "block",
        "blockhash",
        "gasleft",
        "msg",
        "now",
        "tx",
        "this",
        "addmod",
        "mulmod",
        "keccak256",
        "sha256",
        "sha3",
        "ripemd160",
        "ecrecover",
        "selfdestruct",
        "suicide",
        "abi",
        "fallback",
        "receive",
    ]

    # https://solidity.readthedocs.io/en/v0.5.2/miscellaneous.html#reserved-keywords
    RESERVED_KEYWORDS = [
        "abstract",
        "after",
        "alias",
        "apply",
        "auto",
        "case",
        "catch",
        "copyof",
        "default",
        "define",
        "final",
        "immutable",
        "implements",
        "in",
        "inline",
        "let",
        "macro",
        "match",
        "mutable",
        "null",
        "of",
        "override",
        "partial",
        "promise",
        "reference",
        "relocatable",
        "sealed",
        "sizeof",
        "static",
        "supports",
        "switch",
        "try",
        "type",
        "typedef",
        "typeof",
        "unchecked",
    ]

    def is_builtin_symbol(self, word: str | None) -> bool:
        """Detects if a given word is a built-in symbol.

        Returns:
            boolean: True if the given word represents a built-in symbol."""

        return word in self.BUILTIN_SYMBOLS or word in self.RESERVED_KEYWORDS

    def detect_builtin_shadowing_locals(
        self, function_or_modifier: Modifier | FunctionContract
    ) -> list[tuple[str, LocalVariable]]:
        """Detects if local variables in a given function/modifier are named after built-in symbols.
            Any such items are returned in a list.

        Returns:
            list of tuple: (type, definition, local variable parent)"""

        results = []
        for local in function_or_modifier.variables:
            if self.is_builtin_symbol(local.name):
                results.append((self.SHADOWING_LOCAL_VARIABLE, local))
        return results

    def detect_builtin_shadowing_definitions(
        self, contract: Contract
    ) -> list[tuple[str, Function | Variable | Event]]:
        """Detects if functions, access modifiers, events, state variables, or local variables are named after built-in
            symbols. Any such definitions are returned in a list.

        Returns:
            list of tuple: (type, definition, [local variable parent])"""

        result: list[tuple[str, Function | Variable | Event]] = []

        # Loop through all functions, modifiers, variables (state and local) to detect any built-in symbol keywords.
        for function in contract.functions_declared:
            if self.is_builtin_symbol(function.name):
                if function.is_fallback or function.is_receive:
                    continue
                result.append((self.SHADOWING_FUNCTION, function))
            result += self.detect_builtin_shadowing_locals(function)
        for modifier in contract.modifiers_declared:
            if self.is_builtin_symbol(modifier.name):
                result.append((self.SHADOWING_MODIFIER, modifier))
            result += self.detect_builtin_shadowing_locals(modifier)
        for variable in contract.state_variables_declared:
            if self.is_builtin_symbol(variable.name):
                result.append((self.SHADOWING_STATE_VARIABLE, variable))
        for event in contract.events_declared:
            if self.is_builtin_symbol(event.name):
                result.append((self.SHADOWING_EVENT, event))

        return result

    def _detect(self) -> list[Output]:
        """Detect shadowing of built-in symbols

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'shadow'}

        """

        results = []
        for contract in self.contracts:
            shadows = self.detect_builtin_shadowing_definitions(contract)
            if shadows:
                for shadow in shadows:
                    # Obtain components
                    shadow_type = shadow[0]
                    shadow_object = shadow[1]

                    info: DETECTOR_INFO = [
                        shadow_object,
                        f' ({shadow_type}) shadows built-in symbol"\n',
                    ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



"""
Detect if all the given variables are written in all the paths of the function
"""

from collections import defaultdict
from typing import Any

from slither.core.cfg.node import NodeType, Node
from slither.core.declarations import SolidityFunction, Function
from slither.core.variables.variable import Variable
from slither.slithir.operations import (
    Index,
    Member,
    OperationWithLValue,
    SolidityCall,
    Length,
)
from slither.slithir.variables import ReferenceVariable, TemporaryVariable


class State:
    def __init__(self) -> None:
        # Map node -> list of variables set
        # Were each variables set represents a configuration of a path
        # If two paths lead to the exact same set of variables written, we dont need to explore both
        # We need to keep different set per path, because we want to capture stuff like
        # if (..){
        #    v = 10
        # }
        # Here in the endIF node, v can be written, or can be not written. If we were merging the paths
        # We would lose this information
        # In other words, in each in the list represents a set of path that has the same outcome
        self.nodes: dict[Node, list[set[Variable]]] = defaultdict(list)


def _visit(
    node: Node | None,
    state: State,
    variables_written: set[Variable],
    variables_to_write: list[Variable],
) -> list[Variable]:
    """
    Explore all the nodes to look for values not written when the node's function return
    Fixpoint reaches if no new written variables are found

    :param node:
    :param state:
    :param variables_to_write:
    :return:
    """

    refs = {}
    variables_written = set(variables_written)
    if not node:
        return []
    for ir in node.irs:
        if isinstance(ir, SolidityCall):
            # TODO convert the revert to a THROW node
            if ir.function in [
                SolidityFunction("revert(string)"),
                SolidityFunction("revert()"),
            ]:
                return []

        if not isinstance(ir, OperationWithLValue):
            continue
        if isinstance(ir, (Index, Member)):
            refs[ir.lvalue] = ir.variable_left
        if isinstance(ir, Length):
            refs[ir.lvalue] = ir.value

        if ir.lvalue and not isinstance(ir.lvalue, (TemporaryVariable, ReferenceVariable)):
            variables_written.add(ir.lvalue)

        lvalue: Any = ir.lvalue
        while isinstance(lvalue, ReferenceVariable):
            if lvalue not in refs:
                break
            refs_lvalues = refs[lvalue]
            if (
                refs_lvalues
                and isinstance(refs_lvalues, Variable)
                and not isinstance(refs_lvalues, (TemporaryVariable, ReferenceVariable))
            ):
                variables_written.add(refs_lvalues)
            lvalue = refs_lvalues

    ret: list[Variable] = []
    if not node.sons and node.type is not NodeType.THROW:
        ret += [v for v in variables_to_write if v not in variables_written]

    # Explore sons if
    # - Before is none: its the first time we explored the node
    # - variables_written is not before: it means that this path has a configuration of set variables
    # that we haven't seen yet
    before = state.nodes.get(node)
    if before is None or variables_written not in before:
        state.nodes[node].append(variables_written)
        for son in node.sons:
            ret += _visit(son, state, variables_written, variables_to_write)
    return ret


def are_variables_written(function: Function, variables_to_write: list[Variable]) -> list[Variable]:
    """
        Return the list of variable that are not written at the end of the function

    Args:
        function (Function)
        variables_to_write (list Variable): variable that must be written
    Returns:
        list(Variable): List of variable that are not written (sublist of variables_to_write)
    """
    return list(set(_visit(function.entry_point, State(), set(), variables_to_write)))



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Binary, BinaryType
from slither.slithir.variables import Constant
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output
from slither.core.cfg.node import NodeType


class ShiftParameterMixup(AbstractDetector):
    """
    Check for cases where a return(a,b) is used in an assembly function that also returns two variables
    """

    ARGUMENT = "incorrect-shift"
    HELP = "The order of parameters in a shift instruction is incorrect."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-shift-in-assembly"
    )

    WIKI_TITLE = "Incorrect shift in assembly."
    WIKI_DESCRIPTION = "Detect if the values in a shift operation are reversed"

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    function f() internal returns (uint a) {
        assembly {
            a := shr(a, 8)
        }
    }
}
```
The shift statement will right-shift the constant 8 by `a` bits"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Swap the order of parameters."

    def _check_function(self, f: FunctionContract) -> list[Output]:
        results = []
        in_assembly = False

        for node in f.nodes:
            if node.type == NodeType.ASSEMBLY:
                in_assembly = True
                continue
            if node.type == NodeType.ENDASSEMBLY:
                in_assembly = False
                continue
            if not in_assembly:
                continue

            for ir in node.irs:
                if isinstance(ir, Binary) and ir.type in [
                    BinaryType.LEFT_SHIFT,
                    BinaryType.RIGHT_SHIFT,
                ]:
                    if isinstance(ir.variable_left, Constant) and not isinstance(
                        ir.variable_right, Constant
                    ):
                        info: DETECTOR_INFO = [
                            f,
                            " contains an incorrect shift operation: ",
                            node,
                            "\n",
                        ]
                        json = self.generate_result(info)

                        results.append(json)
        return results

    def _detect(self) -> list[Output]:
        results = []
        for c in self.contracts:
            for f in c.functions:
                if f.contract_declarer != c:
                    continue

                if f.contains_assembly:
                    results += self._check_function(f)

        return results



import logging
import sys
import subprocess
from pathlib import Path
import crytic_compile
from slither.tools.mutator.utils.file_handling import create_mutant_file, reset_file
from slither.utils.colors import green, red, yellow

logger = logging.getLogger("Slither-Mutate")


def compile_generated_mutant(file_path: str, mappings: str) -> bool:
    """
    function to compile the generated mutant
    returns: status of compilation
    """
    try:
        crytic_compile.CryticCompile(file_path, solc_remaps=mappings)
        return True
    except Exception:
        return False


def run_test_cmd(
    cmd: str,
    timeout: int | None = None,
    target_file: str | None = None,
    verbose: bool = False,
) -> bool:
    """
    function to run codebase tests
    returns: boolean whether the tests passed or not
    """

    # add --fail-fast for foundry tests, to exit after first failure
    if "forge test" in cmd and "--fail-fast" not in cmd:
        cmd += " --fail-fast"
    # add --bail for hardhat and truffle tests, to exit after first failure
    elif "hardhat test" in cmd or ("truffle test" in cmd and "--bail" not in cmd):
        cmd += " --bail"

    try:
        result = subprocess.run(
            cmd,
            shell=True,
            capture_output=True,
            timeout=timeout,
            check=False,  # True: Raises a CalledProcessError if the return code is non-zero
        )

    except subprocess.TimeoutExpired:
        # Timeout, treat this as a test failure
        logger.info("Tests took too long, consider increasing the timeout")
        result = None  # or set result to a default value

    except KeyboardInterrupt:
        logger.info(yellow("Ctrl-C received"))
        if target_file is not None:
            logger.info("Restoring original files")
            reset_file(target_file)
        logger.info("Exiting")
        sys.exit(1)

    # if result is 0 then it is an uncaught mutant because tests didn't fail
    if result:
        code = result.returncode
        if code == 0:
            return True

    # If tests fail in verbose-mode, print both stdout and stderr for easier debugging
    if verbose:
        logger.info(yellow(result.stdout.decode("utf8")))
        logger.info(red(result.stderr.decode("utf8")))

    return False


# return 0 if uncaught, 1 if caught, and 2 if compilation fails
def test_patch(
    output_folder: Path,
    file: str,
    patch: dict,
    command: str,
    generator_name: str,
    timeout: int,
    mappings: str | None,
    verbose: bool,
) -> int:
    """
    function to verify whether each patch is caught by tests
    returns: 0 (uncaught), 1 (caught), or 2 (compilation failure)
    """
    with open(file, "rb") as filepath:
        content = filepath.read()
    # Perform the replacement based on the index values
    # Note: patch offsets are byte offsets from solc, so we must work with bytes
    replaced_content = (
        content[: patch["start"]] + patch["new_string"].encode("utf8") + content[patch["end"] :]
    )
    # Write the modified content back to the file
    with open(file, "wb") as filepath:
        filepath.write(replaced_content)

    if compile_generated_mutant(file, mappings):
        if run_test_cmd(command, timeout, file, False):
            create_mutant_file(output_folder, file, generator_name)
            logger.info(
                f"[{generator_name}] Line {patch['line_number']}: '{patch['old_string']}' ==> '{patch['new_string']}' --> UNCAUGHT"
            )
            reset_file(file)

            return 0  # uncaught
    else:
        if verbose:
            logger.info(
                yellow(
                    f"[{generator_name}] Line {patch['line_number']}: '{patch['old_string']}' ==> '{patch['new_string']}' --> COMPILATION FAILURE"
                )
            )

        reset_file(file)
        return 2  # compile failure

    if verbose:
        logger.info(
            green(
                f"[{generator_name}] Line {patch['line_number']}: '{patch['old_string']}' ==> '{patch['new_string']}' --> CAUGHT"
            )
        )

    reset_file(file)
    return 1  # caught



from typing import TYPE_CHECKING

from slither.core.declarations.contract_level import ContractLevel
from slither.core.declarations import Event

if TYPE_CHECKING:
    from slither.core.declarations import Contract


class EventContract(Event, ContractLevel):
    def is_declared_by(self, contract: "Contract") -> bool:
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract == contract

    @property
    def canonical_name(self) -> str:
        """Return the function signature as a str
        Returns:
            str: contract.func_name(type1,type2)
        """
        return self.contract.name + "." + self.full_name



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from .arbitrary_send_erc20 import ArbitrarySendErc20


class ArbitrarySendErc20NoPermit(AbstractDetector):
    """
    Detect when `msg.sender` is not used as `from` in transferFrom
    """

    ARGUMENT = "arbitrary-send-erc20"
    HELP = "transferFrom uses arbitrary `from`"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#arbitrary-from-in-transferfrom"

    WIKI_TITLE = "Arbitrary `from` in transferFrom"
    WIKI_DESCRIPTION = "Detect when `msg.sender` is not used as `from` in transferFrom."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    function a(address from, address to, uint256 amount) public {
        erc20.transferFrom(from, to, am);
    }
```
Alice approves this contract to spend her ERC20 tokens. Bob can call `a` and specify Alice's address as the `from` parameter in `transferFrom`, allowing him to transfer Alice's tokens to himself."""

    WIKI_RECOMMENDATION = """
Use `msg.sender` as `from` in transferFrom.
"""

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []

        arbitrary_sends = ArbitrarySendErc20(self.compilation_unit)
        arbitrary_sends.detect()
        for node in arbitrary_sends.no_permit_results:
            func = node.function
            info: DETECTOR_INFO = [func, " uses arbitrary from in transferFrom: ", node, "\n"]
            res = self.generate_result(info)
            results.append(res)

        return results



"""
Module detecting state variables initializing from an immediate function call (prior to constructor run).
"""

from slither.core.declarations.contract import Contract
from slither.core.declarations.function import Function
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from slither.visitors.expression.export_values import ExportValues


def detect_function_init_state_vars(contract: Contract) -> list[StateVariable]:
    """
    Detect any state variables that are initialized from an immediate function call (prior to constructor run).
    :param contract: The contract to detect state variable definitions for.
    :return: A list of all state variables defined in the given contract that meet the specified criteria.
    """
    results = []

    # Loop for each state variable explicitly defined in this contract.
    for state_variable in contract.variables:
        # Skip this variable if it is inherited and not explicitly defined in this contract definition.
        if state_variable.contract != contract:
            continue

        # If it has an expression, we try to break it down to identify if it contains a function call, or reference
        # to a non-constant state variable.
        if state_variable.expression:
            exported_values = ExportValues(state_variable.expression).result()
            for exported_value in exported_values:
                if (
                    isinstance(exported_value, StateVariable) and not exported_value.is_constant
                ) or (isinstance(exported_value, Function) and not exported_value.pure):
                    results.append(state_variable)
                    break

    return results


class FunctionInitializedState(AbstractDetector):
    """
    State variables initializing from an immediate function call (prior to constructor run).
    """

    ARGUMENT = "function-init-state"
    HELP = "Function initializing state variables"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#function-initializing-state"
    )

    WIKI_TITLE = "Function Initializing State"
    WIKI_DESCRIPTION = "Detects the immediate initialization of state variables through function calls that are not pure/constant, or that use non-constant state variable."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract StateVarInitFromFunction {

    uint public v = set(); // Initialize from function (sets to 77)
    uint public w = 5;
    uint public x = set(); // Initialize from function (sets to 88)
    address public shouldntBeReported = address(8);

    constructor(){
        // The constructor is run after all state variables are initialized.
    }

    function set() public  returns(uint)  {
        // If this function is being used to initialize a state variable declared
        // before w, w will be zero. If it is declared after w, w will be set.
        if(w == 0) {
            return 77;
        }

        return 88;
    }
}
```
In this case, users might intend a function to return a value a state variable can initialize with, without realizing the context for the contract is not fully initialized.
In the example above, the same function sets two different values for state variables because it checks a state variable that is not yet initialized in one case, and is initialized in the other.
Special care must be taken when initializing state variables from an immediate function call so as not to incorrectly assume the state is initialized.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove any initialization of state variables via non-constant state variables or function calls. If variables must be set upon contract deployment, locate initialization in the constructor instead."

    def _detect(self) -> list[Output]:
        """
        Detect state variables defined from an immediate function call (pre-contract deployment).

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'shadow'}

        """
        results = []
        for contract in self.contracts:
            state_variables = detect_function_init_state_vars(contract)
            if state_variables:
                for state_variable in state_variables:
                    info: DETECTOR_INFO = [
                        state_variable,
                        " is set pre-construction with a non-constant function or state variable:\n",
                    ]
                    info += [f"\t- {state_variable.expression}\n"]
                    json = self.generate_result(info)
                    results.append(json)

        return results



import re
from slither.formatters.exceptions import FormatImpossible
from slither.formatters.utils.patches import create_patch


# Indicates the recommended versions for replacement
REPLACEMENT_VERSIONS = ["^0.4.25", "^0.5.3"]

# group:
# 0: ^ > >= < <= (optional)
# 1: ' ' (optional)
# 2: version number
# 3: version number
# 4: version number

PATTERN = re.compile(r"(\^|>|>=|<|<=)?([ ]+)?(\d+)\.(\d+)\.(\d+)")


def custom_format(slither, result):
    elements = result["elements"]
    for element in elements:
        solc_version_replace = _determine_solc_version_replacement(
            "".join(element["type_specific_fields"]["directive"][1:])
        )

        _patch(
            slither,
            result,
            element["source_mapping"]["filename_absolute"],
            solc_version_replace,
            element["source_mapping"]["start"],
            element["source_mapping"]["start"] + element["source_mapping"]["length"],
        )


def _determine_solc_version_replacement(used_solc_version):
    versions = PATTERN.findall(used_solc_version)
    if len(versions) == 1:
        version = versions[0]
        minor_version = ".".join(version[2:])[2]
        if minor_version == "4":
            # Replace with 0.4.25
            return "pragma solidity " + REPLACEMENT_VERSIONS[0] + ";"
        if minor_version == "5":
            # Replace with 0.5.3
            return "pragma solidity " + REPLACEMENT_VERSIONS[1] + ";"
        raise FormatImpossible(f"Unknown version {versions}")
    if len(versions) == 2:
        version_right = versions[1]
        minor_version_right = ".".join(version_right[2:])[2]
        if minor_version_right == "4":
            # Replace with 0.4.25
            return "pragma solidity " + REPLACEMENT_VERSIONS[0] + ";"
        if minor_version_right in ["5", "6"]:
            # Replace with 0.5.3
            return "pragma solidity " + REPLACEMENT_VERSIONS[1] + ";"
    return None


def _patch(slither, result, in_file, solc_version, modify_loc_start, modify_loc_end):
    in_file_str = slither.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:modify_loc_end]
    create_patch(
        result,
        in_file,
        int(modify_loc_start),
        int(modify_loc_end),
        old_str_of_interest,
        solc_version,
    )



from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import Function
from slither.core.expressions import BinaryOperation, Identifier, MemberAccess, UnaryOperation
from slither.core.solidity_types import ArrayType
from slither.core.variables import StateVariable
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import Length, Delete, HighLevelCall


class CacheArrayLength(AbstractDetector):
    """
    Detects `for` loops that use `length` member of some storage array in their loop condition and don't modify it.
    """

    ARGUMENT = "cache-array-length"
    HELP = (
        "Detects `for` loops that use `length` member of some storage array in their loop condition and don't "
        "modify it."
    )
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#cache-array-length"

    WIKI_TITLE = "Cache array length"
    WIKI_DESCRIPTION = (
        "Detects `for` loops that use `length` member of some storage array in their loop condition "
        "and don't modify it. "
    )
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C
{
    uint[] array;

    function f() public
    {
        for (uint i = 0; i < array.length; i++)
        {
            // code that does not modify length of `array`
        }
    }
}
```
Since the `for` loop in `f` doesn't modify `array.length`, it is more gas efficient to cache it in some local variable and use that variable instead, like in the following example:

```solidity
contract C
{
    uint[] array;

    function f() public
    {
        uint array_length = array.length;
        for (uint i = 0; i < array_length; i++)
        {
            // code that does not modify length of `array`
        }
    }
}
```
    """
    WIKI_RECOMMENDATION = (
        "Cache the lengths of storage arrays if they are used and not modified in `for` loops."
    )

    @staticmethod
    def _is_identifier_member_access_comparison(exp: BinaryOperation) -> bool:
        """
        Checks whether a BinaryOperation `exp` is an operation on Identifier and MemberAccess.
        """
        return (
            isinstance(exp.expression_left, Identifier)
            and isinstance(exp.expression_right, MemberAccess)
        ) or (
            isinstance(exp.expression_left, MemberAccess)
            and isinstance(exp.expression_right, Identifier)
        )

    @staticmethod
    def _extract_array_from_length_member_access(exp: MemberAccess) -> StateVariable:
        """
        Given a member access `exp`, it returns state array which `length` member is accessed through `exp`.
        If array is not a state array or its `length` member is not referenced, it returns `None`.
        """
        if exp.member_name != "length":
            return None
        if not isinstance(exp.expression, Identifier):
            return None
        if not isinstance(exp.expression.value, StateVariable):
            return None
        if not isinstance(exp.expression.value.type, ArrayType):
            return None
        return exp.expression.value

    @staticmethod
    def _is_loop_referencing_array_length(
        node: Node, visited: set[Node], array: StateVariable, depth: int
    ) -> True:
        """
        For a given loop, checks if it references `array.length` at some point.
        Will also return True if `array.length` is referenced but not changed.
        This may potentially generate false negatives in the detector, but it was done this way because:
        - situations when array `length` is referenced but not modified in loop are rare
        - checking if `array.length` is indeed modified would require much more work
        """
        visited.add(node)
        if node.type == NodeType.STARTLOOP:
            depth += 1
        if node.type == NodeType.ENDLOOP:
            depth -= 1
        if depth == 0:
            return False

        # Array length may change in the following situations:
        # - when `push` is called
        # - when `pop` is called
        # - when `delete` is called on the entire array
        # - when external function call is made (instructions from internal function calls are already in
        #   `node.all_slithir_operations()`, so we don't need to handle internal calls separately)
        if node.type == NodeType.EXPRESSION:
            for op in node.all_slithir_operations():
                if isinstance(op, Length) and op.value == array:
                    # op accesses array.length, not necessarily modifying it
                    return True
                if isinstance(op, Delete):
                    # take into account only delete entire array, since delete array[i] doesn't change `array.length`
                    if (
                        isinstance(op.expression, UnaryOperation)
                        and isinstance(op.expression.expression, Identifier)
                        and op.expression.expression.value == array
                    ):
                        return True
                if (
                    isinstance(op, HighLevelCall)
                    and isinstance(op.function, Function)
                    and not op.function.view
                    and not op.function.pure
                ):
                    return True

        for son in node.sons:
            if son not in visited:
                if CacheArrayLength._is_loop_referencing_array_length(son, visited, array, depth):
                    return True
        return False

    @staticmethod
    def _handle_loops(nodes: list[Node], non_optimal_array_len_usages: list[Node]) -> None:
        """
        For each loop, checks if it has a comparison with `length` array member and, if it has, checks whether that
        array size could potentially change in that loop.
        If it cannot, the loop condition is added to `non_optimal_array_len_usages`.
        There may be some false negatives here - see docs for `_is_loop_referencing_array_length` for more information.
        """
        for node in nodes:
            if node.type == NodeType.STARTLOOP:
                if_node = node.sons[0]
                if if_node.type != NodeType.IFLOOP:
                    continue
                if not isinstance(if_node.expression, BinaryOperation):
                    continue
                exp: BinaryOperation = if_node.expression
                if not CacheArrayLength._is_identifier_member_access_comparison(exp):
                    continue
                array: StateVariable
                if isinstance(exp.expression_right, MemberAccess):
                    array = CacheArrayLength._extract_array_from_length_member_access(
                        exp.expression_right
                    )
                else:  # isinstance(exp.expression_left, MemberAccess) == True
                    array = CacheArrayLength._extract_array_from_length_member_access(
                        exp.expression_left
                    )
                if array is None:
                    continue

                visited: set[Node] = set()
                if not CacheArrayLength._is_loop_referencing_array_length(
                    if_node, visited, array, 1
                ):
                    non_optimal_array_len_usages.append(if_node)

    @staticmethod
    def _get_non_optimal_array_len_usages_for_function(f: Function) -> list[Node]:
        """
        Finds non-optimal usages of array length in loop conditions in a given function.
        """
        non_optimal_array_len_usages: list[Node] = []
        CacheArrayLength._handle_loops(f.nodes, non_optimal_array_len_usages)

        return non_optimal_array_len_usages

    @staticmethod
    def _get_non_optimal_array_len_usages(functions: list[Function]) -> list[Node]:
        """
        Finds non-optimal usages of array length in loop conditions in given functions.
        """
        non_optimal_array_len_usages: list[Node] = []

        for f in functions:
            non_optimal_array_len_usages += (
                CacheArrayLength._get_non_optimal_array_len_usages_for_function(f)
            )

        return non_optimal_array_len_usages

    def _detect(self):
        results = []

        non_optimal_array_len_usages = CacheArrayLength._get_non_optimal_array_len_usages(
            self.compilation_unit.functions
        )
        for usage in non_optimal_array_len_usages:
            info = [
                "Loop condition ",
                usage,
                " should use cached array length instead of referencing `length` member "
                "of the storage array.\n ",
            ]
            res = self.generate_result(info)
            results.append(res)
        return results



import logging

from slither.core.declarations import Contract
from slither.slithir.operations import EventCall
from slither.utils import output

logger = logging.getLogger("Slither-conformance")


def events_safeBatchTransferFrom(contract: Contract, ret: dict[str, list]) -> None:
    function = contract.get_function_from_signature(
        "safeBatchTransferFrom(address,address,uint256[],uint256[],bytes)"
    )
    events = [
        {
            "name": "TransferSingle",
            "parameters": ["address", "address", "address", "uint256", "uint256"],
        },
        {
            "name": "TransferBatch",
            "parameters": ["address", "address", "address", "uint256[]", "uint256[]"],
        },
    ]

    event_counter_name = 0
    event_counter_parameters = 0
    if function:
        for event in events:
            for ir in function.all_slithir_operations():
                if isinstance(ir, EventCall) and ir.name == event["name"]:
                    event_counter_name += 1
                    if event["parameters"] == [str(a.type) for a in ir.arguments]:
                        event_counter_parameters += 1
    if event_counter_parameters == 1 and event_counter_name == 1:
        txt = "[✓] safeBatchTransferFrom emit TransferSingle or TransferBatch"
        logger.info(txt)
    else:
        txt = "[ ] safeBatchTransferFrom must emit TransferSingle or TransferBatch"
        logger.info(txt)

        erroneous_erc1155_safeBatchTransferFrom_event = output.Output(txt)
        erroneous_erc1155_safeBatchTransferFrom_event.add(contract)
        ret["erroneous_erc1155_safeBatchTransferFrom_event"].append(
            erroneous_erc1155_safeBatchTransferFrom_event.data
        )


def check_erc1155(
    contract: Contract, ret: dict[str, list], explored: bool | None = None
) -> dict[str, list]:
    if explored is None:
        explored = set()

    explored.add(contract)

    events_safeBatchTransferFrom(contract, ret)

    for derived_contract in contract.derived_contracts:
        check_erc1155(derived_contract, ret, explored)

    return ret



"""
Module detecting dangerous strict equality

"""

from typing import Any
from slither.analyses.data_dependency.data_dependency import is_dependent_ssa
from slither.core.declarations import Function
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import (
    Assignment,
    Binary,
    BinaryType,
    HighLevelCall,
    SolidityCall,
)

from slither.core.solidity_types import MappingType, ElementaryType

from slither.core.variables.state_variable import StateVariable
from slither.core.declarations.solidity_variables import (
    SolidityVariable,
    SolidityVariableComposed,
    SolidityFunction,
)
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.slithir.operations.operation import Operation
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA
from slither.utils.output import Output
from slither.utils.type import is_underlying_type_address


class IncorrectStrictEquality(AbstractDetector):
    ARGUMENT = "incorrect-equality"
    HELP = "Dangerous strict equalities"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#dangerous-strict-equalities"
    )

    WIKI_TITLE = "Dangerous strict equalities"
    WIKI_DESCRIPTION = "Use of strict equalities that can be easily manipulated by an attacker."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Crowdsale{
    function fund_reached() public returns(bool){
        return this.balance == 100 ether;
    }
```
`Crowdsale` relies on `fund_reached` to know when to stop the sale of tokens.
`Crowdsale` reaches 100 Ether. Bob sends 0.1 Ether. As a result, `fund_reached` is always false and the `crowdsale` never ends."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        """Don't use strict equality to determine if an account has enough Ether or tokens."""
    )

    sources_taint = [
        SolidityVariable("now"),
        SolidityVariableComposed("block.number"),
        SolidityVariableComposed("block.timestamp"),
    ]

    @staticmethod
    def is_direct_comparison(ir: Operation) -> bool:
        return isinstance(ir, Binary) and ir.type == BinaryType.EQUAL

    @staticmethod
    def is_not_comparing_addresses(ir: Binary) -> bool:
        """
        Comparing addresses strictly should not be flagged.
        """

        if is_underlying_type_address(ir.variable_left.type) and is_underlying_type_address(
            ir.variable_right.type
        ):
            return False

        return True

    @staticmethod
    def is_any_tainted(
        variables: list[
            Constant
            | LocalIRVariable
            | TemporaryVariableSSA
            | SolidityVariableComposed
            | SolidityVariable
        ],
        taints: list[
            LocalIRVariable | SolidityVariable | SolidityVariableComposed | TemporaryVariableSSA
        ],
        function: FunctionContract,
    ) -> bool:
        return any(
            is_dependent_ssa(var, taint, function.contract) for var in variables for taint in taints
        )

    def taint_balance_equalities(
        self, functions: list[FunctionContract | Any]
    ) -> list[LocalIRVariable | TemporaryVariableSSA | Any]:
        taints = []
        for func in functions:
            for node in func.nodes:
                for ir in node.irs_ssa:
                    if isinstance(ir, SolidityCall) and ir.function == SolidityFunction(
                        "balance(address)"
                    ):
                        taints.append(ir.lvalue)
                    if isinstance(ir, HighLevelCall):
                        if (
                            isinstance(ir.function, Function)
                            and ir.function.full_name == "balanceOf(address)"
                        ):
                            taints.append(ir.lvalue)
                        if (
                            isinstance(ir.function, StateVariable)
                            and isinstance(ir.function.type, MappingType)
                            and ir.function.name == "balanceOf"
                            and ir.function.type.type_from == ElementaryType("address")
                            and ir.function.type.type_to == ElementaryType("uint256")
                        ):
                            taints.append(ir.lvalue)
                    if isinstance(ir, Assignment):
                        if ir.rvalue in self.sources_taint:
                            taints.append(ir.lvalue)
        return taints

    # Retrieve all tainted (node, function) pairs
    def tainted_equality_nodes(
        self,
        funcs: list[FunctionContract | Any],
        taints: list[LocalIRVariable | TemporaryVariableSSA | Any],
    ) -> dict[FunctionContract, list[Node]]:
        results = {}
        taints += self.sources_taint

        for func in funcs:
            # Disable the detector on top level function until we have good taint on those
            if isinstance(func, FunctionTopLevel):
                continue
            for node in func.nodes:
                for ir in node.irs_ssa:
                    # Filter to only tainted equality (==) comparisons
                    if (
                        self.is_direct_comparison(ir)
                        # Filter out address comparisons which may occur due to lack of field sensitivity in data dependency
                        and self.is_not_comparing_addresses(ir)
                        and self.is_any_tainted(ir.used, taints, func)
                    ):
                        if func not in results:
                            results[func] = []
                        results[func].append(node)

        return results

    def detect_strict_equality(self, contract: Contract) -> dict[FunctionContract, list[Node]]:
        funcs = contract.all_functions_called + contract.modifiers

        # Taint all BALANCE accesses
        taints = self.taint_balance_equalities(funcs)

        # Accumulate tainted (node,function) pairs involved in strict equality (==) comparisons
        results = self.tainted_equality_nodes(funcs, taints)

        return results

    def _detect(self) -> list[Output]:
        results = []

        for c in self.compilation_unit.contracts_derived:
            ret = self.detect_strict_equality(c)

            # sort ret to get deterministic results
            ret = sorted(ret.items(), key=lambda x: x[0].name)
            for f, nodes in ret:
                func_info = [f, " uses a dangerous strict equality:\n"]

                # sort the nodes to get deterministic results
                nodes.sort(key=lambda x: x.node_id)

                # Output each node with the function info header as a separate result.
                for node in nodes:
                    node_info = func_info + ["\t- ", node, "\n"]

                    res = self.generate_result(node_info)
                    results.append(res)

        return results



import argparse
import logging
import sys
from typing import Any

from crytic_compile import cryticparser

from slither import Slither
from slither.tools.properties.properties.erc20 import generate_erc20, ERC20_PROPERTIES
from slither.tools.properties.addresses.address import (
    Addresses,
    OWNER_ADDRESS,
    USER_ADDRESS,
    ATTACKER_ADDRESS,
)
from slither.utils.myprettytable import MyPrettyTable

logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)

logger = logging.getLogger("Slither")
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(message)s")
logger.addHandler(ch)
logger.handlers[0].setFormatter(formatter)
logger.propagate = False


def _all_scenarios() -> str:
    txt = "\n"
    txt += "#################### ERC20 ####################\n"
    for k, value in ERC20_PROPERTIES.items():
        txt += f"{k} - {value.description}\n"

    return txt


def _all_properties() -> MyPrettyTable:
    table = MyPrettyTable(["Num", "Description", "Scenario"])
    idx = 0
    for scenario, value in ERC20_PROPERTIES.items():
        for prop in value.properties:
            table.add_row([str(idx), prop.description, scenario])
            idx = idx + 1
    return table


class ListScenarios(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        logger.info(_all_scenarios())
        parser.exit()


class ListProperties(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        logger.info(_all_properties())
        parser.exit()


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Generates code properties (e.g., invariants) that can be tested with unit tests or Echidna, entirely automatically.",
        usage="slither-prop filename",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "filename", help="The filename of the contract or project directory to analyze."
    )

    parser.add_argument("--contract", help="The targeted contract.")

    parser.add_argument(
        "--scenario",
        help="Test a specific scenario. Use --list-scenarios to see the available scenarios. Default Transferable",
        default="Transferable",
    )

    parser.add_argument(
        "--list-scenarios",
        help="List available scenarios",
        action=ListScenarios,
        nargs=0,
        default=False,
    )

    parser.add_argument(
        "--list-properties",
        help="List available properties",
        action=ListProperties,
        nargs=0,
        default=False,
    )

    parser.add_argument(
        "--address-owner", help=f"Owner address. Default {OWNER_ADDRESS}", default=None
    )

    parser.add_argument(
        "--address-user", help=f"Owner address. Default {USER_ADDRESS}", default=None
    )

    parser.add_argument(
        "--address-attacker",
        help=f"Attacker address. Default {ATTACKER_ADDRESS}",
        default=None,
    )

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    # Perform slither analysis on the given filename
    slither = Slither(args.filename, **vars(args))

    contracts = slither.get_contract_from_name(args.contract)
    if len(contracts) != 1:
        if len(slither.contracts) == 1:
            contract = slither.contracts[0]
        else:
            if args.contract is None:
                to_log = "Specify the target: --contract ContractName"
            else:
                to_log = f"{args.contract} not found"
            logger.error(to_log)
            return
    else:
        contract = contracts[0]

    addresses = Addresses(args.address_owner, args.address_user, args.address_attacker)

    generate_erc20(contract, args.scenario, addresses)


if __name__ == "__main__":
    main()



from collections import defaultdict

from crytic_compile.platform import Type as PlatformType

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


def _find_missing_inheritance(compilation_unit: SlitherCompilationUnit) -> list[Contract]:
    """
    Filter contracts with missing inheritance to return only the "most base" contracts
    in the inheritance tree.
    :param slither:
    :return:
    """
    missings = compilation_unit.contracts_with_missing_inheritance

    ret = []
    for b in missings:
        is_most_base = True
        for inheritance in b.immediate_inheritance:
            if inheritance in missings:
                is_most_base = False
        if is_most_base:
            ret.append(b)

    return ret


class NameReused(AbstractDetector):
    ARGUMENT = "name-reused"
    HELP = "Contract's name reused"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#name-reused"

    WIKI_TITLE = "Name reused"

    # region wiki_description
    WIKI_DESCRIPTION = """If a codebase has two contracts the similar names, the compilation artifacts
will not contain one of the contracts with the duplicate name."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
Bob's `truffle` codebase has two contracts named `ERC20`.
When `truffle compile` runs, only one of the two contracts will generate artifacts in `build/contracts`.
As a result, the second contract cannot be analyzed.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Rename the contract."

    def _detect(self) -> list[Output]:
        results = []
        compilation_unit = self.compilation_unit
        if compilation_unit.core.crytic_compile.platform != PlatformType.TRUFFLE:
            return []

        all_contracts = compilation_unit.contracts
        all_contracts_name = [c.name for c in all_contracts]
        contracts_name_reused = {
            contract for contract in all_contracts_name if all_contracts_name.count(contract) > 1
        }

        names_reused = {
            name: compilation_unit.get_contract_from_name(name) for name in contracts_name_reused
        }

        # First show the contracts that we know are missing
        incorrectly_constructed = [
            contract
            for contract in compilation_unit.contracts
            if contract.is_incorrectly_constructed
        ]

        inheritance_corrupted = defaultdict(list)
        for contract in incorrectly_constructed:
            for father in contract.inheritance:
                inheritance_corrupted[father.name].append(contract)

        for contract_name, files in names_reused.items():
            info: DETECTOR_INFO = [contract_name, " is re-used:\n"]
            for file in files:
                if file is None:
                    info += ["\t- In an file not found, most likely in\n"]
                else:
                    info += ["\t- ", file, "\n"]

            if contract_name in inheritance_corrupted:
                info += ["\tAs a result, the inherited contracts are not correctly analyzed:\n"]
            for corrupted in inheritance_corrupted[contract_name]:
                info += ["\t\t- ", corrupted, "\n"]
            res = self.generate_result(info)
            results.append(res)

        # Then show the contracts for which one of the father was not found
        # Here we are not able to know
        most_base_with_missing_inheritance = _find_missing_inheritance(compilation_unit)

        for b in most_base_with_missing_inheritance:
            info = [b, " inherits from a contract for which the name is reused.\n"]
            if b.inheritance:
                info += ["\t- Slither could not determine which contract has a duplicate name:\n"]
                for inheritance in b.inheritance:
                    info += ["\t\t-", inheritance, "\n"]
                info += ["\t- Check if:\n"]
                info += ["\t\t- A inherited contract is missing from this list,\n"]
                info += ["\t\t- The contract are imported from the correct files.\n"]
            if b.derived_contracts:
                info += [f"\t- This issue impacts the contracts inheriting from {b.name}:\n"]
                for derived in b.derived_contracts:
                    info += ["\t\t-", derived, "\n"]
            res = self.generate_result(info)
            results.append(res)
        return results



from pathlib import Path
from typing import TYPE_CHECKING
from hashlib import sha1
from slither.utils.oz_hashes import oz_hashes

if TYPE_CHECKING:
    from slither.core.declarations import Contract

libraries = {
    "Openzeppelin-SafeMath": lambda x: is_openzeppelin_safemath(x),
    "Openzeppelin-ECRecovery": lambda x: is_openzeppelin_ecrecovery(x),
    "Openzeppelin-Ownable": lambda x: is_openzeppelin_ownable(x),
    "Openzeppelin-ERC20": lambda x: is_openzeppelin_erc20(x),
    "Openzeppelin-ERC721": lambda x: is_openzeppelin_erc721(x),
    "Zos-Upgrade": lambda x: is_zos_initializable(x),
    "Dapphub-DSAuth": lambda x: is_dapphub_ds_auth(x),
    "Dapphub-DSMath": lambda x: is_dapphub_ds_math(x),
    "Dapphub-DSToken": lambda x: is_dapphub_ds_token(x),
    "Dapphub-DSProxy": lambda x: is_dapphub_ds_proxy(x),
    "Dapphub-DSGroup": lambda x: is_dapphub_ds_group(x),
    "AragonOS-SafeMath": lambda x: is_aragonos_safemath(x),
    "AragonOS-ERC20": lambda x: is_aragonos_erc20(x),
    "AragonOS-AppProxyBase": lambda x: is_aragonos_app_proxy_base(x),
    "AragonOS-AppProxyPinned": lambda x: is_aragonos_app_proxy_pinned(x),
    "AragonOS-AppProxyUpgradeable": lambda x: is_aragonos_app_proxy_upgradeable(x),
    "AragonOS-AppStorage": lambda x: is_aragonos_app_storage(x),
    "AragonOS-AragonApp": lambda x: is_aragonos_aragon_app(x),
    "AragonOS-UnsafeAragonApp": lambda x: is_aragonos_unsafe_aragon_app(x),
    "AragonOS-Autopetrified": lambda x: is_aragonos_autopetrified(x),
    "AragonOS-DelegateProxy": lambda x: is_aragonos_delegate_proxy(x),
    "AragonOS-DepositableDelegateProxy": lambda x: is_aragonos_depositable_delegate_proxy(x),
    "AragonOS-DepositableStorage": lambda x: is_aragonos_delegate_proxy(x),
    "AragonOS-Initializable": lambda x: is_aragonos_initializable(x),
    "AragonOS-IsContract": lambda x: is_aragonos_is_contract(x),
    "AragonOS-Petrifiable": lambda x: is_aragonos_petrifiable(x),
    "AragonOS-ReentrancyGuard": lambda x: is_aragonos_reentrancy_guard(x),
    "AragonOS-TimeHelpers": lambda x: is_aragonos_time_helpers(x),
    "AragonOS-VaultRecoverable": lambda x: is_aragonos_vault_recoverable(x),
}


def is_standard_library(contract: "Contract") -> str | None:
    for name, is_lib in libraries.items():
        if is_lib(contract):
            return name
    return None


###################################################################################
###################################################################################
# region General libraries
###################################################################################
###################################################################################


def is_openzeppelin(contract: "Contract") -> bool:
    if not contract.is_from_dependency():
        return False
    path = Path(contract.source_mapping.filename.absolute).parts
    is_zep = "openzeppelin-solidity" in Path(contract.source_mapping.filename.absolute).parts
    try:
        is_zep |= path[path.index("@openzeppelin") + 1] == "contracts"
    except IndexError:
        pass
    except ValueError:
        pass
    return is_zep


def is_openzeppelin_strict(contract: "Contract") -> bool:
    source_hash = sha1(contract.source_mapping.content.encode("utf8")).hexdigest()
    return source_hash in oz_hashes


def is_zos(contract: "Contract") -> bool:
    if not contract.is_from_dependency():
        return False
    return "zos-lib" in Path(contract.source_mapping.filename.absolute).parts


def is_aragonos(contract: "Contract") -> bool:
    if not contract.is_from_dependency():
        return False
    return "@aragon/os" in Path(contract.source_mapping.filename.absolute).parts


# endregion
###################################################################################
###################################################################################
# region SafeMath
###################################################################################
###################################################################################


def is_safemath(contract: "Contract") -> bool:
    return contract.name == "SafeMath"


def is_openzeppelin_safemath(contract: "Contract") -> bool:
    return is_safemath(contract) and is_openzeppelin(contract)


def is_aragonos_safemath(contract: "Contract") -> bool:
    return is_safemath(contract) and is_aragonos(contract)


# endregion
###################################################################################
###################################################################################
# region ECRecovery
###################################################################################
###################################################################################


def is_ecrecovery(contract: "Contract") -> bool:
    return contract.name == "ECRecovery"


def is_openzeppelin_ecrecovery(contract: "Contract") -> bool:
    return is_ecrecovery(contract) and is_openzeppelin(contract)


# endregion
###################################################################################
###################################################################################
# region Ownable
###################################################################################
###################################################################################


def is_ownable(contract: "Contract") -> bool:
    return contract.name == "Ownable"


def is_openzeppelin_ownable(contract: "Contract") -> bool:
    return is_ownable(contract) and is_openzeppelin(contract)


# endregion
###################################################################################
###################################################################################
# region ERC20
###################################################################################
###################################################################################


def is_erc20(contract: "Contract") -> bool:
    return contract.name == "ERC20"


def is_openzeppelin_erc20(contract: "Contract") -> bool:
    return is_erc20(contract) and is_openzeppelin(contract)


def is_aragonos_erc20(contract: "Contract") -> bool:
    return is_erc20(contract) and is_aragonos(contract)


# endregion
###################################################################################
###################################################################################
# region ERC721
###################################################################################
###################################################################################


def is_erc721(contract: "Contract") -> bool:
    return contract.name == "ERC721"


def is_openzeppelin_erc721(contract: "Contract") -> bool:
    return is_erc721(contract) and is_openzeppelin(contract)


# endregion
###################################################################################
###################################################################################
# region Zos Initializable
###################################################################################
###################################################################################


def is_initializable(contract: "Contract") -> bool:
    return contract.name == "Initializable"


def is_zos_initializable(contract: "Contract") -> bool:
    return is_initializable(contract) and is_zos(contract)


# endregion
###################################################################################
###################################################################################
# region DappHub
###################################################################################
###################################################################################

dapphubs = {
    "DSAuth": "ds-auth",
    "DSMath": "ds-math",
    "DSToken": "ds-token",
    "DSProxy": "ds-proxy",
    "DSGroup": "ds-group",
}


def _is_ds(contract: "Contract", name: str) -> bool:
    return contract.name == name


def _is_dappdhub_ds(contract: "Contract", name: str) -> bool:
    if not contract.is_from_dependency():
        return False
    if dapphubs[name] not in Path(contract.source_mapping.filename.absolute).parts:
        return False
    return _is_ds(contract, name)


def is_ds_auth(contract: "Contract") -> bool:
    return _is_ds(contract, "DSAuth")


def is_dapphub_ds_auth(contract: "Contract") -> bool:
    return _is_dappdhub_ds(contract, "DSAuth")


def is_ds_math(contract: "Contract") -> bool:
    return _is_ds(contract, "DSMath")


def is_dapphub_ds_math(contract: "Contract") -> bool:
    return _is_dappdhub_ds(contract, "DSMath")


def is_ds_token(contract: "Contract") -> bool:
    return _is_ds(contract, "DSToken")


def is_dapphub_ds_token(contract: "Contract") -> bool:
    return _is_dappdhub_ds(contract, "DSToken")


def is_ds_proxy(contract: "Contract") -> bool:
    return _is_ds(contract, "DSProxy")


def is_dapphub_ds_proxy(contract: "Contract") -> bool:
    return _is_dappdhub_ds(contract, "DSProxy")


def is_ds_group(contract: "Contract") -> bool:
    return _is_ds(contract, "DSGroup")


def is_dapphub_ds_group(contract: "Contract") -> bool:
    return _is_dappdhub_ds(contract, "DSGroup")


# endregion
###################################################################################
###################################################################################
# region Aragon
###################################################################################
###################################################################################


def is_aragonos_app_proxy_base(contract: "Contract") -> bool:
    return contract.name == "AppProxyBase" and is_aragonos(contract)


def is_aragonos_app_proxy_pinned(contract: "Contract") -> bool:
    return contract.name == "AppProxyPinned" and is_aragonos(contract)


def is_aragonos_app_proxy_upgradeable(contract: "Contract") -> bool:
    return contract.name == "AppProxyUpgradeable" and is_aragonos(contract)


def is_aragonos_app_storage(contract: "Contract") -> bool:
    return contract.name == "AppStorage" and is_aragonos(contract)


def is_aragonos_aragon_app(contract: "Contract") -> bool:
    return contract.name == "AragonApp" and is_aragonos(contract)


def is_aragonos_unsafe_aragon_app(contract: "Contract") -> bool:
    return contract.name == "UnsafeAragonApp" and is_aragonos(contract)


def is_aragonos_autopetrified(contract: "Contract") -> bool:
    return contract.name == "Autopetrified" and is_aragonos(contract)


def is_aragonos_delegate_proxy(contract: "Contract") -> bool:
    return contract.name == "DelegateProxy" and is_aragonos(contract)


def is_aragonos_depositable_delegate_proxy(contract: "Contract") -> bool:
    return contract.name == "DepositableDelegateProxy" and is_aragonos(contract)


def is_aragonos_depositable_storage(contract: "Contract") -> bool:
    return contract.name == "DepositableStorage" and is_aragonos(contract)


def is_aragonos_ether_token_contract(contract: "Contract") -> bool:
    return contract.name == "EtherTokenConstant" and is_aragonos(contract)


def is_aragonos_initializable(contract: "Contract") -> bool:
    return contract.name == "Initializable" and is_aragonos(contract)


def is_aragonos_is_contract(contract: "Contract") -> bool:
    return contract.name == "IsContract" and is_aragonos(contract)


def is_aragonos_petrifiable(contract: "Contract") -> bool:
    return contract.name == "Petrifiable" and is_aragonos(contract)


def is_aragonos_reentrancy_guard(contract: "Contract") -> bool:
    return contract.name == "ReentrancyGuard" and is_aragonos(contract)


def is_aragonos_time_helpers(contract: "Contract") -> bool:
    return contract.name == "TimeHelpers" and is_aragonos(contract)


def is_aragonos_vault_recoverable(contract: "Contract") -> bool:
    return contract.name == "VaultRecoverable" and is_aragonos(contract)



from pathlib import Path

from slither.tools.properties.addresses.address import Addresses
from slither.tools.properties.utils import write_file


def generate_echidna_config(output_dir: Path, addresses: Addresses) -> str:
    """
    Generate the echidna configuration file
    :param output_dir:
    :param addresses:
    :return:
    """
    content = "prefix: crytic_\n"
    content += f'deployer: "{addresses.owner}"\n'
    content += f'sender: ["{addresses.user}", "{addresses.attacker}"]\n'
    content += f'psender: "{addresses.user}"\n'
    content += "coverage: true\n"
    filename = "echidna_config.yaml"
    write_file(output_dir, filename, content)
    return filename



"""
Module detecting incorrect operator usage for exponentiation where bitwise xor '^' is used instead of '**'
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Contract, Function
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Binary, BinaryType, Operation
from slither.slithir.utils.utils import RVALUE
from slither.slithir.variables.constant import Constant
from slither.utils.output import Output


def _is_constant_candidate(var: RVALUE | Function) -> bool:
    """
    Check if the variable is a constant.
    Do not consider variable that are expressed with hexadecimal.
    Something like 2^0xf is likely to be a correct bitwise operator
    :param var:
    :return:
    """
    return isinstance(var, Constant) and not var.original_value.startswith("0x")


def _is_bitwise_xor_on_constant(ir: Operation) -> bool:
    return (
        isinstance(ir, Binary)
        and ir.type == BinaryType.CARET
        and (_is_constant_candidate(ir.variable_left) or _is_constant_candidate(ir.variable_right))
    )


def _detect_incorrect_operator(contract: Contract) -> list[tuple[Function, Node]]:
    ret: list[tuple[Function, Node]] = []
    f: Function
    for f in contract.functions + contract.modifiers:  # type:ignore
        # Heuristic: look for binary expressions with ^ operator where at least one of the operands is a constant, and
        # the constant is not in hex, because hex typically is used with bitwise xor and not exponentiation
        nodes = [node for node in f.nodes for ir in node.irs if _is_bitwise_xor_on_constant(ir)]
        for node in nodes:
            ret.append((f, node))
    return ret


class IncorrectOperatorExponentiation(AbstractDetector):
    """
    Incorrect operator usage of bitwise xor mistaking it for exponentiation
    """

    ARGUMENT = "incorrect-exp"
    HELP = "Incorrect exponentiation"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-exponentiation"

    WIKI_TITLE = "Incorrect exponentiation"
    WIKI_DESCRIPTION = "Detect use of bitwise `xor ^` instead of exponential `**`"
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Bug{
    uint UINT_MAX = 2^256 - 1;
    ...
}
```
Alice deploys a contract in which `UINT_MAX` incorrectly uses `^` operator instead of `**` for exponentiation"""

    WIKI_RECOMMENDATION = "Use the correct operator `**` for exponentiation."

    def _detect(self) -> list[Output]:
        """Detect the incorrect operator usage for exponentiation where bitwise xor ^ is used instead of **

        Returns:
            list: (function, node)
        """
        results: list[Output] = []
        for c in self.compilation_unit.contracts_derived:
            res = _detect_incorrect_operator(c)
            for func, node in res:
                info: DETECTOR_INFO = [
                    func,
                    " has bitwise-xor operator ^ instead of the exponentiation operator **: \n",
                ]
                info += ["\t - ", node, "\n"]
                results.append(self.generate_result(info))

        return results



""" "
Contract module
"""

import logging
from collections import defaultdict
from pathlib import Path
from typing import Optional, TYPE_CHECKING, Union, Any
from collections.abc import Callable

from crytic_compile.platform import Type as PlatformType

from slither.core.cfg.scope import Scope
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.utils.using_for import USING_FOR, merge_using_for
from slither.core.declarations.function import Function, FunctionType, FunctionLanguage
from slither.utils.erc import (
    ERC20_signatures,
    ERC165_signatures,
    ERC223_signatures,
    ERC721_signatures,
    ERC1820_signatures,
    ERC777_signatures,
    ERC1155_signatures,
    ERC2612_signatures,
    ERC1363_signatures,
    ERC4524_signatures,
    ERC4626_signatures,
)
from slither.utils.tests_pattern import is_test_contract


if TYPE_CHECKING:
    from slither.core.declarations import (
        Enum,
        EventContract,
        Modifier,
        EnumContract,
        StructureContract,
        FunctionContract,
        CustomErrorContract,
    )
    from slither.slithir.operations import HighLevelCall, LibraryCall
    from slither.slithir.variables.variable import SlithIRVariable
    from slither.core.variables import Variable, StateVariable
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.scope.scope import FileScope
    from slither.core.cfg.node import Node
    from slither.core.solidity_types import TypeAliasContract


LOGGER = logging.getLogger("Contract")


class Contract(SourceMapping):
    """
    Contract class
    """

    def __init__(self, compilation_unit: "SlitherCompilationUnit", scope: "FileScope") -> None:
        super().__init__()

        self._name: str | None = None
        self._id: int | None = None
        self._inheritance: list[Contract] = []  # all contract inherited, c3 linearization
        self._immediate_inheritance: list[Contract] = []  # immediate inheritance
        # Start slot for the persistent storage if custom layout used
        self._custom_storage_layout: int | None = None

        # Constructors called on contract's definition
        # contract B is A(1) { ..
        self._explicit_base_constructor_calls: list[Contract] = []

        self._enums: dict[str, EnumContract] = {}
        self._structures: dict[str, StructureContract] = {}
        self._events: dict[str, EventContract] = {}
        # map accessible variable from name -> variable
        # do not contain private variables inherited from contract
        self._variables: dict[str, StateVariable] = {}
        self._variables_ordered: list[StateVariable] = []
        # Reference id -> variable declaration (only available for compact AST)
        self._state_variables_by_ref_id: dict[int, StateVariable] = {}
        self._modifiers: dict[str, Modifier] = {}
        self._functions: dict[str, FunctionContract] = {}
        self._linearizedBaseContracts: list[int] = []
        self._custom_errors: dict[str, CustomErrorContract] = {}
        self._type_aliases: dict[str, TypeAliasContract] = {}

        # The only str is "*"
        self._using_for: USING_FOR = {}
        self._using_for_complete: USING_FOR | None = None
        self._kind: str | None = None
        self._is_interface: bool = False
        self._is_library: bool = False
        self._is_fully_implemented: bool = False
        self._is_abstract: bool = False

        self._signatures: list[str] | None = None
        self._signatures_declared: list[str] | None = None

        self._fallback_function: FunctionContract | None = None
        self._receive_function: FunctionContract | None = None

        self._is_upgradeable: bool | None = None
        self._is_upgradeable_proxy: bool | None = None
        self._upgradeable_version: str | None = None

        self._initial_state_variables: list[StateVariable] = []  # ssa

        self._is_incorrectly_parsed: bool = False

        self._available_functions_as_dict: dict[str, Function] | None = None
        self._all_functions_called: list[Function] | None = None

        self.compilation_unit: SlitherCompilationUnit = compilation_unit
        self.file_scope: FileScope = scope

        # memoize
        self._state_variables_used_in_reentrant_targets: (
            dict[StateVariable, set[StateVariable | Function]] | None
        ) = None

        self._comments: str | None = None

    ###################################################################################
    ###################################################################################
    # region General's properties
    ###################################################################################
    ###################################################################################

    @property
    def name(self) -> str:
        """str: Name of the contract."""
        assert self._name
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def id(self) -> int:
        """Unique id."""
        assert self._id is not None
        return self._id

    @id.setter
    def id(self, new_id: int) -> None:
        """Unique id."""
        self._id = new_id

    @property
    def contract_kind(self) -> str | None:
        """
        contract_kind can be None if the legacy ast format is used
        :return:
        """
        return self._kind

    @contract_kind.setter
    def contract_kind(self, kind: str) -> None:
        self._kind = kind

    @property
    def is_interface(self) -> bool:
        return self._is_interface

    @is_interface.setter
    def is_interface(self, is_interface: bool) -> None:
        self._is_interface = is_interface

    @property
    def is_library(self) -> bool:
        return self._is_library

    @is_library.setter
    def is_library(self, is_library: bool) -> None:
        self._is_library = is_library

    @property
    def comments(self) -> str | None:
        """
        Return the comments associated with the contract.

        When using comments, avoid strict text matching, as the solc behavior might change.
        For example, for old solc version, the first space after the * is not kept, i.e:

          * @title Test Contract
          * @dev Test comment

        Returns
        - " @title Test Contract\n @dev Test comment" for newest versions
        - "@title Test Contract\n@dev Test comment" for older versions


        Returns:
            the comment as a string
        """
        return self._comments

    @comments.setter
    def comments(self, comments: str):
        self._comments = comments

    @property
    def is_fully_implemented(self) -> bool:
        """
        bool: True if the contract defines all functions.
        In modern Solidity, virtual functions can lack an implementation.
        Prior to Solidity 0.6.0, functions like the following would be not fully implemented:
        ```solidity
        contract ImplicitAbstract{
            function f() public;
        }
        ```
        """
        return self._is_fully_implemented

    @is_fully_implemented.setter
    def is_fully_implemented(self, is_fully_implemented: bool):
        self._is_fully_implemented = is_fully_implemented

    @property
    def is_abstract(self) -> bool:
        """
        Note for Solidity < 0.6.0 it will always be false
        bool: True if the contract is abstract.
        """
        return self._is_abstract

    @is_abstract.setter
    def is_abstract(self, is_abstract: bool):
        self._is_abstract = is_abstract

    @property
    def custom_storage_layout(self) -> int | None:
        """
        Return the persistent storage slot starting position if a custom storage layout is used.
        Otherwise None.
        int: Storage slot starting position.
        """
        return self._custom_storage_layout

    @custom_storage_layout.setter
    def custom_storage_layout(self, slot: int):
        self._custom_storage_layout = slot

    # endregion
    ###################################################################################
    ###################################################################################
    # region Structures
    ###################################################################################
    ###################################################################################

    @property
    def structures(self) -> list["StructureContract"]:
        """
        list(Structure): List of the structures
        """
        return list(self._structures.values())

    @property
    def structures_inherited(self) -> list["StructureContract"]:
        """
        list(Structure): List of the inherited structures
        """
        return [s for s in self.structures if s.contract != self]

    @property
    def structures_declared(self) -> list["StructureContract"]:
        """
        list(Structues): List of the structures declared within the contract (not inherited)
        """
        return [s for s in self.structures if s.contract == self]

    @property
    def structures_as_dict(self) -> dict[str, "StructureContract"]:
        return self._structures

    # endregion
    ###################################################################################
    ###################################################################################
    # region Enums
    ###################################################################################
    ###################################################################################

    @property
    def enums(self) -> list["EnumContract"]:
        return list(self._enums.values())

    @property
    def enums_inherited(self) -> list["EnumContract"]:
        """
        list(Enum): List of the inherited enums
        """
        return [e for e in self.enums if e.contract != self]

    @property
    def enums_declared(self) -> list["EnumContract"]:
        """
        list(Enum): List of the enums declared within the contract (not inherited)
        """
        return [e for e in self.enums if e.contract == self]

    @property
    def enums_as_dict(self) -> dict[str, "EnumContract"]:
        return self._enums

    # endregion
    ###################################################################################
    ###################################################################################
    # region Events
    ###################################################################################
    ###################################################################################

    @property
    def events(self) -> list["EventContract"]:
        """
        list(Event): List of the events
        """
        return list(self._events.values())

    @property
    def events_inherited(self) -> list["EventContract"]:
        """
        list(Event): List of the inherited events
        """
        return [e for e in self.events if e.contract != self]

    @property
    def events_declared(self) -> list["EventContract"]:
        """
        list(Event): List of the events declared within the contract (not inherited)
        """
        return [e for e in self.events if e.contract == self]

    @property
    def events_as_dict(self) -> dict[str, "EventContract"]:
        return self._events

    # endregion
    ###################################################################################
    ###################################################################################
    # region Using for
    ###################################################################################
    ###################################################################################

    @property
    def using_for(self) -> USING_FOR:
        return self._using_for

    @property
    def using_for_complete(self) -> USING_FOR:
        """
        USING_FOR: Dict of merged local using for directive with top level directive
        """

        if self._using_for_complete is None:
            result = self.using_for
            top_level_using_for = self.file_scope.using_for_directives
            for uftl in top_level_using_for:
                result = merge_using_for(result, uftl.using_for)
            self._using_for_complete = result
        return self._using_for_complete

    # endregion
    ###################################################################################
    ###################################################################################
    # region Custom Errors
    ###################################################################################
    ###################################################################################

    @property
    def custom_errors(self) -> list["CustomErrorContract"]:
        """
        list(CustomErrorContract): List of the contract's custom errors
        """
        return list(self._custom_errors.values())

    @property
    def custom_errors_inherited(self) -> list["CustomErrorContract"]:
        """
        list(CustomErrorContract): List of the inherited custom errors
        """
        return [s for s in self.custom_errors if s.contract != self]

    @property
    def custom_errors_declared(self) -> list["CustomErrorContract"]:
        """
        list(CustomErrorContract): List of the custom errors declared within the contract (not inherited)
        """
        return [s for s in self.custom_errors if s.contract == self]

    @property
    def custom_errors_as_dict(self) -> dict[str, "CustomErrorContract"]:
        return self._custom_errors

    # endregion
    ###################################################################################
    ###################################################################################
    # region Custom Errors
    ###################################################################################
    ###################################################################################

    @property
    def type_aliases(self) -> list["TypeAliasContract"]:
        """
        list(TypeAliasContract): List of the contract's custom errors
        """
        return list(self._type_aliases.values())

    @property
    def type_aliases_inherited(self) -> list["TypeAliasContract"]:
        """
        list(TypeAliasContract): List of the inherited custom errors
        """
        return [s for s in self.type_aliases if s.contract != self]

    @property
    def type_aliases_declared(self) -> list["TypeAliasContract"]:
        """
        list(TypeAliasContract): List of the custom errors declared within the contract (not inherited)
        """
        return [s for s in self.type_aliases if s.contract == self]

    @property
    def type_aliases_as_dict(self) -> dict[str, "TypeAliasContract"]:
        return self._type_aliases

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################
    @property
    def state_variables_by_ref_id(self) -> dict[int, "StateVariable"]:
        """
        Returns the state variables by reference id (only available for compact AST).
        """
        return self._state_variables_by_ref_id

    @property
    def variables(self) -> list["StateVariable"]:
        """
        Returns all the accessible variables (do not include private variable from inherited contract)

        list(StateVariable): List of the state variables. Alias to self.state_variables.
        """
        return list(self.state_variables)

    @property
    def variables_as_dict(self) -> dict[str, "StateVariable"]:
        return self._variables

    @property
    def state_variables(self) -> list["StateVariable"]:
        """
        Returns all the accessible variables (do not include private variable from inherited contract).
        Use stored_state_variables_ordered for all the storage variables following the storage order
        Use transient_state_variables_ordered for all the transient variables following the storage order

        list(StateVariable): List of the state variables.
        """
        return list(self._variables.values())

    @property
    def state_variables_entry_points(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the state variables that are public.
        """
        return [var for var in self._variables.values() if var.visibility == "public"]

    @property
    def state_variables_ordered(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the state variables by order of declaration.
        """
        return self._variables_ordered

    def add_state_variables_ordered(self, new_vars: list["StateVariable"]) -> None:
        self._variables_ordered += new_vars

    @property
    def storage_variables_ordered(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the state variables in storage location by order of declaration.
        """
        return [v for v in self._variables_ordered if v.is_stored]

    @property
    def transient_variables_ordered(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the state variables in transient location by order of declaration.
        """
        return [v for v in self._variables_ordered if v.is_transient]

    @property
    def state_variables_inherited(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the inherited state variables
        """
        return [s for s in self.state_variables if s.contract != self]

    @property
    def state_variables_declared(self) -> list["StateVariable"]:
        """
        list(StateVariable): List of the state variables declared within the contract (not inherited)
        """
        return [s for s in self.state_variables if s.contract == self]

    @property
    def slithir_variables(self) -> list["SlithIRVariable"]:
        """
        List all of the slithir variables (non SSA)
        """
        slithir_variabless = [f.slithir_variables for f in self.functions + self.modifiers]  # type: ignore
        slithir_variables = [item for sublist in slithir_variabless for item in sublist]
        return list(set(slithir_variables))

    @property
    def state_variables_used_in_reentrant_targets(
        self,
    ) -> dict["StateVariable", set[Union["StateVariable", "Function"]]]:
        """
        Returns the state variables used in reentrant targets. Heuristics:
        - Variable used (read/write) in entry points that are reentrant
        - State variables that are public

        """
        from slither.core.variables.state_variable import StateVariable

        if self._state_variables_used_in_reentrant_targets is None:
            reentrant_functions = [f for f in self.functions_entry_points if f.is_reentrant]
            variables_used: dict[StateVariable, set[StateVariable | Function]] = defaultdict(set)
            for function in reentrant_functions:
                for ir in function.all_slithir_operations():
                    state_variables = [v for v in ir.used if isinstance(v, StateVariable)]
                    for state_variable in state_variables:
                        variables_used[state_variable].add(ir.node.function)
            for variable in [v for v in self.state_variables if v.visibility == "public"]:
                variables_used[variable].add(variable)
            self._state_variables_used_in_reentrant_targets = variables_used
        return self._state_variables_used_in_reentrant_targets

    # endregion
    ###################################################################################
    ###################################################################################
    # region Constructors
    ###################################################################################
    ###################################################################################

    @property
    def constructor(self) -> Optional["Function"]:
        """
        Return the contract's immediate constructor.
        If there is no immediate constructor, returns the first constructor
        executed, following the c3 linearization
        Return None if there is no constructor.
        """
        cst = self.constructors_declared
        if cst:
            return cst
        for inherited_contract in self.inheritance:
            cst = inherited_contract.constructors_declared
            if cst:
                return cst
        return None

    @property
    def constructors_declared(self) -> Optional["Function"]:
        return next(
            (
                func
                for func in self.functions
                if func.is_constructor and func.contract_declarer == self
            ),
            None,
        )

    @property
    def constructors(self) -> list["FunctionContract"]:
        """
        Return the list of constructors (including inherited)
        """
        return [func for func in self.functions if func.is_constructor]

    @property
    def explicit_base_constructor_calls(self) -> list["Function"]:
        """
        list(Function): List of the base constructors called explicitly by this contract definition.

                        Base constructors called by any constructor definition will not be included.
                        Base constructors implicitly called by the contract definition (without
                        parenthesis) will not be included.

                        On "contract B is A(){..}" it returns the constructor of A
        """
        return [c.constructor for c in self._explicit_base_constructor_calls if c.constructor]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions and Modifiers
    ###################################################################################
    ###################################################################################

    @property
    def functions_signatures(self) -> list[str]:
        """
        Return the signatures of all the public/eterxnal functions/state variables
        :return: list(string) the signatures of all the functions that can be called
        """
        if self._signatures is None:
            sigs = [
                v.full_name for v in self.state_variables if v.visibility in ["public", "external"]
            ]

            sigs += {f.full_name for f in self.functions if f.visibility in ["public", "external"]}
            self._signatures = list(set(sigs))
        return self._signatures

    @property
    def functions_signatures_declared(self) -> list[str]:
        """
        Return the signatures of the public/eterxnal functions/state variables that are declared by this contract
        :return: list(string) the signatures of all the functions that can be called and are declared by this contract
        """
        if self._signatures_declared is None:
            sigs = [
                v.full_name
                for v in self.state_variables_declared
                if v.visibility in ["public", "external"]
            ]

            sigs += {
                f.full_name
                for f in self.functions_declared
                if f.visibility in ["public", "external"]
            }
            self._signatures_declared = list(set(sigs))
        return self._signatures_declared

    @property
    def functions(self) -> list["FunctionContract"]:
        """
        list(Function): List of the functions
        """
        return list(self._functions.values())

    def available_functions_as_dict(self) -> dict[str, "Function"]:
        if self._available_functions_as_dict is None:
            self._available_functions_as_dict = {
                f.full_name: f for f in self._functions.values() if not f.is_shadowed
            }
        return self._available_functions_as_dict

    def add_function(self, func: "FunctionContract") -> None:
        self._functions[func.canonical_name] = func

    def set_functions(self, functions: dict[str, "FunctionContract"]) -> None:
        """
        Set the functions

        :param functions:  dict full_name -> function
        :return:
        """
        self._functions = functions

    @property
    def functions_inherited(self) -> list["FunctionContract"]:
        """
        list(Function): List of the inherited functions
        """
        return [f for f in self.functions if f.contract_declarer != self]

    @property
    def functions_declared(self) -> list["FunctionContract"]:
        """
        list(Function): List of the functions defined within the contract (not inherited)
        """
        return [f for f in self.functions if f.contract_declarer == self]

    @property
    def functions_entry_points(self) -> list["FunctionContract"]:
        """
        list(Functions): List of public and external functions
        """
        return [
            f
            for f in self.functions
            if (f.visibility in ["public", "external"] and not f.is_shadowed) or f.is_fallback
        ]

    @property
    def modifiers(self) -> list["Modifier"]:
        """
        list(Modifier): List of the modifiers
        """
        return list(self._modifiers.values())

    def available_modifiers_as_dict(self) -> dict[str, "Modifier"]:
        return {m.full_name: m for m in self._modifiers.values() if not m.is_shadowed}

    def set_modifiers(self, modifiers: dict[str, "Modifier"]) -> None:
        """
        Set the modifiers

        :param modifiers:  dict full_name -> modifier
        :return:
        """
        self._modifiers = modifiers

    @property
    def modifiers_inherited(self) -> list["Modifier"]:
        """
        list(Modifier): List of the inherited modifiers
        """
        return [m for m in self.modifiers if m.contract_declarer != self]

    @property
    def modifiers_declared(self) -> list["Modifier"]:
        """
        list(Modifier): List of the modifiers defined within the contract (not inherited)
        """
        return [m for m in self.modifiers if m.contract_declarer == self]

    @property
    def functions_and_modifiers(self) -> list["Function"]:
        """
        list(Function|Modifier): List of the functions and modifiers
        """
        return self.functions + self.modifiers  # type: ignore

    @property
    def functions_and_modifiers_inherited(self) -> list["Function"]:
        """
        list(Function|Modifier): List of the inherited functions and modifiers
        """
        return self.functions_inherited + self.modifiers_inherited  # type: ignore

    @property
    def functions_and_modifiers_declared(self) -> list["Function"]:
        """
        list(Function|Modifier): List of the functions and modifiers defined within the contract (not inherited)
        """
        return self.functions_declared + self.modifiers_declared  # type: ignore

    @property
    def fallback_function(self) -> Optional["FunctionContract"]:
        if self._fallback_function is None:
            for f in self.functions:
                if f.is_fallback:
                    self._fallback_function = f
                    break
        return self._fallback_function

    @property
    def receive_function(self) -> Optional["FunctionContract"]:
        if self._receive_function is None:
            for f in self.functions:
                if f.is_receive:
                    self._receive_function = f
                    break
        return self._receive_function

    def available_elements_from_inheritances(
        self,
        elements: dict[str, "Function"],
        getter_available: Callable[["Contract"], list["FunctionContract"]],
    ) -> dict[str, "Function"]:
        """

        :param elements: dict(canonical_name -> elements)
        :param getter_available: fun x
        :return:
        """
        # keep track of the contracts visited
        # to prevent an ovveride due to multiple inheritance of the same contract
        # A is B, C, D is C, --> the second C was already seen
        inherited_elements: dict[str, FunctionContract] = {}
        accessible_elements = {}
        contracts_visited = []
        for father in self.inheritance_reverse:
            functions: dict[str, FunctionContract] = {
                v.full_name: v
                for v in getter_available(father)
                if v.contract not in contracts_visited
                and v.function_language
                != FunctionLanguage.Yul  # Yul functions are not propagated in the inheritance
            }
            contracts_visited.append(father)
            inherited_elements.update(functions)

        for element in inherited_elements.values():
            accessible_elements[element.full_name] = elements[element.canonical_name]

        return accessible_elements

    # endregion
    ###################################################################################
    ###################################################################################
    # region Inheritance
    ###################################################################################
    ###################################################################################

    @property
    def inheritance(self) -> list["Contract"]:
        """
        list(Contract): Inheritance list. Order: the first elem is the first father to be executed
        """
        return list(self._inheritance)

    @property
    def immediate_inheritance(self) -> list["Contract"]:
        """
        list(Contract): List of contracts immediately inherited from (fathers). Order: order of declaration.
        """
        return list(self._immediate_inheritance)

    @property
    def inheritance_reverse(self) -> list["Contract"]:
        """
        list(Contract): Inheritance list. Order: the last elem is the first father to be executed
        """
        return list(reversed(self._inheritance))

    def set_inheritance(
        self,
        inheritance: list["Contract"],
        immediate_inheritance: list["Contract"],
        called_base_constructor_contracts: list["Contract"],
    ) -> None:
        self._inheritance = inheritance
        self._immediate_inheritance = immediate_inheritance
        self._explicit_base_constructor_calls = called_base_constructor_contracts

    @property
    def derived_contracts(self) -> list["Contract"]:
        """
        list(Contract): Return the list of contracts derived from self
        """
        candidates = self.compilation_unit.contracts
        return [c for c in candidates if self in c.inheritance]  # type: ignore

    # endregion
    ###################################################################################
    ###################################################################################
    # region Getters from/to object
    ###################################################################################
    ###################################################################################

    def get_functions_reading_from_variable(self, variable: "Variable") -> list["Function"]:
        """
        Return the functions reading the variable
        """
        return [f for f in self.functions if f.is_reading(variable)]

    def get_functions_writing_to_variable(self, variable: "Variable") -> list["Function"]:
        """
        Return the functions writing the variable
        """
        return [f for f in self.functions if f.is_writing(variable)]

    def get_function_from_full_name(self, full_name: str) -> Optional["Function"]:
        """
            Return a function from a full name
            The full name differs from the solidity's signature are the type are conserved
            For example contract type are kept, structure are not unrolled, etc
        Args:
            full_name (str): signature of the function (without return statement)
        Returns:
            Function
        """
        return next(
            (f for f in self.functions if f.full_name == full_name and not f.is_shadowed),
            None,
        )

    def get_function_from_signature(self, function_signature: str) -> Optional["Function"]:
        """
            Return a function from a signature
        Args:
            function_signature (str): signature of the function (without return statement)
        Returns:
            Function
        """
        return next(
            (
                f
                for f in self.functions
                if f.solidity_signature == function_signature and not f.is_shadowed
            ),
            None,
        )

    def get_modifier_from_signature(self, modifier_signature: str) -> Optional["Modifier"]:
        """
        Return a modifier from a signature

        :param modifier_signature:
        """
        return next(
            (m for m in self.modifiers if m.full_name == modifier_signature and not m.is_shadowed),
            None,
        )

    def get_function_from_canonical_name(self, canonical_name: str) -> Optional["Function"]:
        """
            Return a function from a canonical name (contract.signature())
        Args:
            canonical_name (str): canonical name of the function (without return statement)
        Returns:
            Function
        """
        return next((f for f in self.functions if f.canonical_name == canonical_name), None)

    def get_modifier_from_canonical_name(self, canonical_name: str) -> Optional["Modifier"]:
        """
            Return a modifier from a canonical name (contract.signature())
        Args:
            canonical_name (str): canonical name of the modifier
        Returns:
            Modifier
        """
        return next((m for m in self.modifiers if m.canonical_name == canonical_name), None)

    def get_state_variable_from_name(self, variable_name: str) -> Optional["StateVariable"]:
        """
        Return a state variable from a name

        :param variable_name:
        """
        return next((v for v in self.state_variables if v.name == variable_name), None)

    def get_state_variable_from_canonical_name(
        self, canonical_name: str
    ) -> Optional["StateVariable"]:
        """
            Return a state variable from a canonical_name
        Args:
            canonical_name (str): name of the variable
        Returns:
            StateVariable
        """
        return next((v for v in self.state_variables if v.canonical_name == canonical_name), None)

    def get_structure_from_name(self, structure_name: str) -> Optional["StructureContract"]:
        """
            Return a structure from a name
        Args:
            structure_name (str): name of the structure
        Returns:
            StructureContract
        """
        return next((st for st in self.structures if st.name == structure_name), None)

    def get_structure_from_canonical_name(
        self, structure_name: str
    ) -> Optional["StructureContract"]:
        """
            Return a structure from a canonical name
        Args:
            structure_name (str): canonical name of the structure
        Returns:
            StructureContract
        """
        return next((st for st in self.structures if st.canonical_name == structure_name), None)

    def get_event_from_signature(self, event_signature: str) -> Optional["Event"]:
        """
            Return an event from a signature
        Args:
            event_signature (str): signature of the event
        Returns:
            Event
        """
        return next((e for e in self.events if e.full_name == event_signature), None)

    def get_event_from_canonical_name(self, event_canonical_name: str) -> Optional["Event"]:
        """
            Return an event from a canonical name
        Args:
            event_canonical_name (str): name of the event
        Returns:
            Event
        """
        return next((e for e in self.events if e.canonical_name == event_canonical_name), None)

    def get_enum_from_name(self, enum_name: str) -> Optional["Enum"]:
        """
            Return an enum from a name
        Args:
            enum_name (str): name of the enum
        Returns:
            Enum
        """
        return next((e for e in self.enums if e.name == enum_name), None)

    def get_enum_from_canonical_name(self, enum_name: str) -> Optional["Enum"]:
        """
            Return an enum from a canonical name
        Args:
            enum_name (str): canonical name of the enum
        Returns:
            Enum
        """
        return next((e for e in self.enums if e.canonical_name == enum_name), None)

    def get_functions_overridden_by(self, function: "Function") -> list["Function"]:
        """
            Return the list of functions overridden by the function
        Args:
            (core.Function)
        Returns:
            list(core.Function)

        """
        return function.overrides

    # endregion
    ###################################################################################
    ###################################################################################
    # region Recursive getters
    ###################################################################################
    ###################################################################################

    @property
    def all_functions_called(self) -> list["Function"]:
        """
        list(Function): List of functions reachable from the contract
        Includes super, and private/internal functions not shadowed
        """
        from slither.slithir.operations import Operation

        if self._all_functions_called is None:
            all_functions = [f for f in self.functions + self.modifiers if not f.is_shadowed]  # type: ignore
            all_callss = [f.all_internal_calls() for f in all_functions] + [list(all_functions)]
            all_calls = [
                item.function if isinstance(item, Operation) else item
                for sublist in all_callss
                for item in sublist
            ]
            all_calls = list(set(all_calls))

            all_constructors = [c.constructor for c in self.inheritance if c.constructor]
            all_constructors = list(set(all_constructors))

            set_all_calls = set(all_calls + list(all_constructors))

            self._all_functions_called = [c for c in set_all_calls if isinstance(c, Function)]
        return self._all_functions_called

    @property
    def all_state_variables_written(self) -> list["StateVariable"]:
        """
        list(StateVariable): List all of the state variables written
        """
        all_state_variables_writtens = [
            f.all_state_variables_written()
            for f in self.functions + self.modifiers  # type: ignore
        ]
        all_state_variables_written = [
            item for sublist in all_state_variables_writtens for item in sublist
        ]
        return list(set(all_state_variables_written))

    @property
    def all_state_variables_read(self) -> list["StateVariable"]:
        """
        list(StateVariable): List all of the state variables read
        """
        all_state_variables_reads = [
            f.all_state_variables_read()
            for f in self.functions + self.modifiers  # type: ignore
        ]
        all_state_variables_read = [
            item for sublist in all_state_variables_reads for item in sublist
        ]
        return list(set(all_state_variables_read))

    @property
    def all_library_calls(self) -> list["LibraryCall"]:
        """
        list(LibraryCall): List all of the libraries func called
        """
        all_high_level_callss = [f.all_library_calls() for f in self.functions + self.modifiers]  # type: ignore
        all_high_level_calls = [item for sublist in all_high_level_callss for item in sublist]
        return list(set(all_high_level_calls))

    @property
    def all_high_level_calls(self) -> list[tuple["Contract", "HighLevelCall"]]:
        """
        list(Tuple("Contract", "HighLevelCall")): List all of the external high level calls
        """
        all_high_level_callss = [f.all_high_level_calls() for f in self.functions + self.modifiers]  # type: ignore
        all_high_level_calls = [item for sublist in all_high_level_callss for item in sublist]
        return list(set(all_high_level_calls))

    # endregion
    ###################################################################################
    ###################################################################################
    # region Summary information
    ###################################################################################
    ###################################################################################

    def get_summary(
        self, include_shadowed: bool = True
    ) -> tuple[str, list[str], list[str], list, list]:
        """Return the function summary

        :param include_shadowed: boolean to indicate if shadowed functions should be included (default True)
        Returns:
            (str, list, list, list, list): (name, inheritance, variables, fuction summaries, modifier summaries)
        """
        func_summaries = [
            f.get_summary() for f in self.functions if (not f.is_shadowed or include_shadowed)
        ]
        modif_summaries = [
            f.get_summary() for f in self.modifiers if (not f.is_shadowed or include_shadowed)
        ]
        return (
            self.name,
            [str(x) for x in self.inheritance],
            [str(x) for x in self.variables],
            func_summaries,
            modif_summaries,
        )

    def is_signature_only(self) -> bool:
        """Detect if the contract has only abstract functions

        Returns:
            bool: true if the function are abstract functions
        """
        return all((not f.is_implemented) for f in self.functions)

    # endregion
    ###################################################################################
    ###################################################################################
    # region ERC conformance
    ###################################################################################
    ###################################################################################

    def ercs(self) -> list[str]:
        """
        Return the ERC implemented
        :return: list of string
        """
        all_erc = [
            ("ERC20", self.is_erc20),
            ("ERC165", self.is_erc165),
            ("ERC1820", self.is_erc1820),
            ("ERC223", self.is_erc223),
            ("ERC721", self.is_erc721),
            ("ERC777", self.is_erc777),
            ("ERC2612", self.is_erc2612),
            ("ERC1363", self.is_erc1363),
            ("ERC4626", self.is_erc4626),
        ]

        return [erc for erc, is_erc in all_erc if is_erc()]

    def is_erc20(self) -> bool:
        """
            Check if the contract is an erc20 token

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc20
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC20_signatures)

    def is_erc165(self) -> bool:
        """
            Check if the contract is an erc165 token

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc165
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC165_signatures)

    def is_erc1820(self) -> bool:
        """
            Check if the contract is an erc1820

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc165
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC1820_signatures)

    def is_erc223(self) -> bool:
        """
            Check if the contract is an erc223 token

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc223
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC223_signatures)

    def is_erc721(self) -> bool:
        """
            Check if the contract is an erc721 token

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc721
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC721_signatures)

    def is_erc777(self) -> bool:
        """
            Check if the contract is an erc777

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc165
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC777_signatures)

    def is_erc1155(self) -> bool:
        """
            Check if the contract is an erc1155

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc1155
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC1155_signatures)

    def is_erc4626(self) -> bool:
        """
            Check if the contract is an erc4626

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc4626
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC4626_signatures)

    def is_erc2612(self) -> bool:
        """
            Check if the contract is an erc2612

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc2612
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC2612_signatures)

    def is_erc1363(self) -> bool:
        """
            Check if the contract is an erc1363

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc1363
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC1363_signatures)

    def is_erc4524(self) -> bool:
        """
            Check if the contract is an erc4524

            Note: it does not check for correct return values
        :return: Returns a true if the contract is an erc4524
        """
        full_names = self.functions_signatures
        return all(s in full_names for s in ERC4524_signatures)

    @property
    def is_token(self) -> bool:
        """
        Check if the contract follows one of the standard ERC token
        :return:
        """
        return (
            self.is_erc20()
            or self.is_erc721()
            or self.is_erc165()
            or self.is_erc223()
            or self.is_erc777()
            or self.is_erc1155()
        )

    def is_possible_erc20(self) -> bool:
        """
        Checks if the provided contract could be attempting to implement ERC20 standards.

        :return: Returns a boolean indicating if the provided contract met the token standard.
        """
        # We do not check for all the functions, as name(), symbol(), might give too many FPs
        full_names = self.functions_signatures
        return (
            "transfer(address,uint256)" in full_names
            or "transferFrom(address,address,uint256)" in full_names
            or "approve(address,uint256)" in full_names
        )

    def is_possible_erc721(self) -> bool:
        """
        Checks if the provided contract could be attempting to implement ERC721 standards.

        :return: Returns a boolean indicating if the provided contract met the token standard.
        """
        # We do not check for all the functions, as name(), symbol(), might give too many FPs
        full_names = self.functions_signatures
        return (
            "ownerOf(uint256)" in full_names
            or "safeTransferFrom(address,address,uint256,bytes)" in full_names
            or "safeTransferFrom(address,address,uint256)" in full_names
            or "setApprovalForAll(address,bool)" in full_names
            or "getApproved(uint256)" in full_names
            or "isApprovedForAll(address,address)" in full_names
        )

    @property
    def is_possible_token(self) -> bool:
        """
        Check if the contract is a potential token (it might not implement all the functions)
        :return:
        """
        return self.is_possible_erc20() or self.is_possible_erc721()

    # endregion
    ###################################################################################
    ###################################################################################
    # region Dependencies
    ###################################################################################
    ###################################################################################

    def is_from_dependency(self) -> bool:
        return self.compilation_unit.core.crytic_compile.is_dependency(
            self.source_mapping.filename.absolute
        )

    # endregion
    ###################################################################################
    ###################################################################################
    # region Test
    ###################################################################################
    ###################################################################################

    @property
    def is_truffle_migration(self) -> bool:
        """
        Return true if the contract is the Migrations contract needed for Truffle
        :return:
        """
        if self.compilation_unit.core.crytic_compile.platform == PlatformType.TRUFFLE:
            if self.name == "Migrations":
                paths = Path(self.source_mapping.filename.absolute).parts
                if len(paths) >= 2:
                    return paths[-2] == "contracts" and paths[-1] == "migrations.sol"
        return False

    @property
    def is_test(self) -> bool:
        return is_test_contract(self) or self.is_truffle_migration  # type: ignore

    # endregion
    ###################################################################################
    ###################################################################################
    # region Function analyses
    ###################################################################################
    ###################################################################################

    def update_read_write_using_ssa(self) -> None:
        for function in self.functions + list(self.modifiers):
            function.update_read_write_using_ssa()

    # endregion
    ###################################################################################
    ###################################################################################
    # region Upgradeability
    ###################################################################################
    ###################################################################################

    @property
    def is_upgradeable(self) -> bool:
        if self._is_upgradeable is None:
            self._is_upgradeable = False
            initializable = self.file_scope.get_contract_from_name("Initializable")
            if initializable:
                if initializable in self.inheritance:
                    self._is_upgradeable = True
            else:
                for contract in self.inheritance + [self]:
                    # This might lead to false positive
                    lower_name = contract.name.lower()
                    if "upgradeable" in lower_name or "upgradable" in lower_name:
                        self._is_upgradeable = True
                        break
                    if "initializable" in lower_name:
                        self._is_upgradeable = True
                        break
        return self._is_upgradeable

    @is_upgradeable.setter
    def is_upgradeable(self, upgradeable: bool) -> None:
        self._is_upgradeable = upgradeable

    @property
    def is_upgradeable_proxy(self) -> bool:
        from slither.core.cfg.node import NodeType
        from slither.slithir.operations import LowLevelCall

        if self._is_upgradeable_proxy is None:
            self._is_upgradeable_proxy = False
            if "Proxy" in self.name:
                self._is_upgradeable_proxy = True
                return True
            for f in self.functions:
                if f.is_fallback:
                    for node in f.all_nodes():
                        for ir in node.irs:
                            if isinstance(ir, LowLevelCall) and ir.function_name == "delegatecall":
                                self._is_upgradeable_proxy = True
                                return self._is_upgradeable_proxy
                        if node.type == NodeType.ASSEMBLY:
                            inline_asm = node.inline_asm
                            if inline_asm:
                                if "delegatecall" in inline_asm:
                                    self._is_upgradeable_proxy = True
                                    return self._is_upgradeable_proxy
        return self._is_upgradeable_proxy

    @is_upgradeable_proxy.setter
    def is_upgradeable_proxy(self, upgradeable_proxy: bool) -> None:
        self._is_upgradeable_proxy = upgradeable_proxy

    @property
    def upgradeable_version(self) -> str | None:
        return self._upgradeable_version

    @upgradeable_version.setter
    def upgradeable_version(self, version_name: str) -> None:
        self._upgradeable_version = version_name

    # endregion
    ###################################################################################
    ###################################################################################
    # region Internals
    ###################################################################################
    ###################################################################################

    @property
    def is_incorrectly_constructed(self) -> bool:
        """
        Return true if there was an internal Slither's issue when analyzing the contract
        :return:
        """
        return self._is_incorrectly_parsed

    @is_incorrectly_constructed.setter
    def is_incorrectly_constructed(self, incorrect: bool) -> None:
        self._is_incorrectly_parsed = incorrect

    def add_constructor_variables(self) -> None:
        from slither.core.declarations.function_contract import FunctionContract

        if self.state_variables:
            for idx, variable_candidate in enumerate(self.state_variables):
                if variable_candidate.expression and not variable_candidate.is_constant:
                    constructor_variable = FunctionContract(self.compilation_unit)
                    constructor_variable.set_function_type(FunctionType.CONSTRUCTOR_VARIABLES)
                    constructor_variable.set_contract(self)  # type: ignore
                    constructor_variable.set_contract_declarer(self)  # type: ignore
                    constructor_variable.set_visibility("internal")
                    # For now, source mapping of the constructor variable is the whole contract
                    # Could be improved with a targeted source mapping
                    constructor_variable.set_offset(self.source_mapping, self.compilation_unit)
                    self._functions[constructor_variable.canonical_name] = constructor_variable

                    prev_node = self._create_node(
                        constructor_variable, 0, variable_candidate, constructor_variable
                    )
                    variable_candidate.node_initialization = prev_node
                    counter = 1
                    for v in self.state_variables[idx + 1 :]:
                        if v.expression and not v.is_constant:
                            next_node = self._create_node(
                                constructor_variable, counter, v, prev_node.scope
                            )
                            v.node_initialization = next_node
                            prev_node.add_son(next_node)
                            next_node.add_father(prev_node)
                            prev_node = next_node
                            counter += 1
                    break

            for idx, variable_candidate in enumerate(self.state_variables):
                if variable_candidate.expression and variable_candidate.is_constant:
                    constructor_variable = FunctionContract(self.compilation_unit)
                    constructor_variable.set_function_type(
                        FunctionType.CONSTRUCTOR_CONSTANT_VARIABLES
                    )
                    constructor_variable.set_contract(self)  # type: ignore
                    constructor_variable.set_contract_declarer(self)  # type: ignore
                    constructor_variable.set_visibility("internal")
                    # For now, source mapping of the constructor variable is the whole contract
                    # Could be improved with a targeted source mapping
                    constructor_variable.set_offset(self.source_mapping, self.compilation_unit)
                    self._functions[constructor_variable.canonical_name] = constructor_variable

                    prev_node = self._create_node(
                        constructor_variable, 0, variable_candidate, constructor_variable
                    )
                    variable_candidate.node_initialization = prev_node
                    counter = 1
                    for v in self.state_variables[idx + 1 :]:
                        if v.expression and v.is_constant:
                            next_node = self._create_node(
                                constructor_variable, counter, v, prev_node.scope
                            )
                            v.node_initialization = next_node
                            prev_node.add_son(next_node)
                            next_node.add_father(prev_node)
                            prev_node = next_node
                            counter += 1

                    break

    def _create_node(
        self, func: Function, counter: int, variable: "Variable", scope: Scope | Function
    ) -> "Node":
        from slither.core.cfg.node import Node, NodeType
        from slither.core.expressions import (
            AssignmentOperationType,
            AssignmentOperation,
            Identifier,
        )

        # Function uses to create node for state variable declaration statements
        node = Node(NodeType.OTHER_ENTRYPOINT, counter, scope, func.file_scope)
        node.set_offset(variable.source_mapping, self.compilation_unit)
        node.set_function(func)
        func.add_node(node)
        assert variable.expression
        expression = AssignmentOperation(
            Identifier(variable),
            variable.expression,
            AssignmentOperationType.ASSIGN,
            variable.type,
        )

        expression.set_offset(variable.source_mapping, self.compilation_unit)
        node.add_expression(expression)
        return node

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIR
    ###################################################################################
    ###################################################################################

    def convert_expression_to_slithir_ssa(self) -> None:
        """
        Assume generate_slithir_and_analyze was called on all functions

        :return:
        """
        from slither.slithir.variables import StateIRVariable

        all_ssa_state_variables_instances = {}

        for contract in self.inheritance:
            for v in contract.state_variables_declared:
                new_var = StateIRVariable(v)
                all_ssa_state_variables_instances[v.canonical_name] = new_var
                self._initial_state_variables.append(new_var)

        for v in self.variables:
            if v.contract == self:
                new_var = StateIRVariable(v)
                all_ssa_state_variables_instances[v.canonical_name] = new_var
                self._initial_state_variables.append(new_var)

        for func in self.functions + list(self.modifiers):
            func.generate_slithir_ssa(all_ssa_state_variables_instances)

    def fix_phi(self) -> None:
        last_state_variables_instances: dict[str, list[StateVariable]] = {}
        initial_state_variables_instances: dict[str, StateVariable] = {}
        for v in self._initial_state_variables:
            last_state_variables_instances[v.canonical_name] = []
            initial_state_variables_instances[v.canonical_name] = v

        for func in self.functions + list(self.modifiers):
            result = func.get_last_ssa_state_variables_instances()
            for variable_name, instances in result.items():
                # TODO: investigate the next operation
                last_state_variables_instances[variable_name] += list(instances)

        for func in self.functions + list(self.modifiers):
            func.fix_phi(last_state_variables_instances, initial_state_variables_instances)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Built in definitions
    ###################################################################################
    ###################################################################################

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return other == self.name
        return NotImplemented

    def __neq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return other != self.name
        return NotImplemented

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return self._id  # type:ignore

    # endregion
